# Licensed under a 3-clause BSD style license - see LICENSE.rst
import operator
import logging
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.table import Table
import matplotlib.pyplot as plt
from gammapy.maps import MapAxis, RegionNDMap
from gammapy.modeling.models import (
    integrate_spectrum,
    SPECTRAL_MODEL_REGISTRY,
    BrokenPowerLawSpectralModel,
    CompoundSpectralModel,
    ConstantSpectralModel,
    EBLAbsorptionNormSpectralModel,
    ExpCutoffPowerLaw3FGLSpectralModel,
    ExpCutoffPowerLawNormSpectralModel,
    ExpCutoffPowerLawSpectralModel,
    GaussianSpectralModel,
    LogParabolaNormSpectralModel,
    LogParabolaSpectralModel,
    Model,
    NaimaSpectralModel,
    PiecewiseNormSpectralModel,
    PowerLaw2SpectralModel,
    PowerLawNormSpectralModel,
    PowerLawSpectralModel,
    SkyModel,
    SmoothBrokenPowerLawSpectralModel,
    SuperExpCutoffPowerLaw4FGLDR3SpectralModel,
    SuperExpCutoffPowerLaw4FGLSpectralModel,
    TemplateNDSpectralModel,
    TemplateSpectralModel,
)
from gammapy.utils.compat import COPY_IF_NEEDED
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import (
    assert_quantity_allclose,
    mpl_plot_check,
    requires_data,
    requires_dependency,
)
from gammapy.utils.deprecation import GammapyDeprecationWarning


def table_model():
    energy = MapAxis.from_energy_bounds(0.1 * u.TeV, 100 * u.TeV, 1000).center

    model = PowerLawSpectralModel(
        index=2.3, amplitude="4 cm-2 s-1 TeV-1", reference="1 TeV"
    )
    dnde = model(energy)

    return TemplateSpectralModel(energy, dnde)


TEST_MODELS = [
    dict(
        name="constant",
        model=ConstantSpectralModel(const=4 / u.cm**2 / u.s / u.TeV),
        val_at_2TeV=u.Quantity(4, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(35.9999999999999, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(198.00000000000006, "TeV cm-2 s-1"),
    ),
    dict(
        name="powerlaw",
        model=PowerLawSpectralModel(
            index=2.3 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(4 * 2.0 ** (-2.3), "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.9227116204223784, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(6.650836884969039, "TeV cm-2 s-1"),
    ),
    dict(
        name="powerlaw",
        model=PowerLawSpectralModel(
            index=2 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(1.0, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(3.6, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(9.210340371976184, "TeV cm-2 s-1"),
    ),
    dict(
        name="norm-powerlaw",
        model=PowerLawNormSpectralModel(
            tilt=2 * u.Unit(""),
            norm=4.0 * u.Unit(""),
            reference=1 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(1.0, ""),
        integral_1_10TeV=u.Quantity(3.6, "TeV"),
        eflux_1_10TeV=u.Quantity(9.210340371976184, "TeV2"),
    ),
    dict(
        name="powerlaw2",
        model=PowerLaw2SpectralModel(
            amplitude=u.Quantity(2.9227116204223784, "cm-2 s-1"),
            index=2.3 * u.Unit(""),
            emin=1 * u.TeV,
            emax=10 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(4 * 2.0 ** (-2.3), "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.9227116204223784, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(6.650836884969039, "TeV cm-2 s-1"),
    ),
    dict(
        name="ecpl",
        model=ExpCutoffPowerLawSpectralModel(
            index=1.6 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            lambda_=0.1 / u.TeV,
        ),
        val_at_2TeV=u.Quantity(1.080321705479446, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(3.765838739678921, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(9.901735870666526, "TeV cm-2 s-1"),
        e_peak=4 * u.TeV,
    ),
    dict(
        name="norm-ecpl",
        model=ExpCutoffPowerLawNormSpectralModel(
            index=1.6 * u.Unit(""),
            norm=4 * u.Unit(""),
            reference=1 * u.TeV,
            lambda_=0.1 / u.TeV,
        ),
        val_at_2TeV=u.Quantity(1.080321705479446, ""),
        integral_1_10TeV=u.Quantity(3.765838739678921, "TeV"),
        eflux_1_10TeV=u.Quantity(9.901735870666526, "TeV2"),
    ),
    dict(
        name="ecpl_3fgl",
        model=ExpCutoffPowerLaw3FGLSpectralModel(
            index=2.3 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            ecut=10 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(0.7349563611124971, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.6034046173089, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(5.340285560055799, "TeV cm-2 s-1"),
    ),
    dict(
        name="plsec_4fgl_dr1",
        model=SuperExpCutoffPowerLaw4FGLSpectralModel(
            index_1=1.5,
            index_2=2,
            amplitude=1 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            expfactor=1e-14,
        ),
        val_at_2TeV=u.Quantity(0.3431043087721737, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(1.2125247, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(3.38072082, "TeV cm-2 s-1"),
    ),
    dict(
        name="plsec_4fgl",
        model=SuperExpCutoffPowerLaw4FGLDR3SpectralModel(
            index_1=1.5,
            index_2=2,
            amplitude=1 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            expfactor=1e-2,
        ),
        val_at_2TeV=u.Quantity(0.35212994, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(1.328499, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(4.067067, "TeV cm-2 s-1"),
        e_peak=10.0498756 * u.TeV,
    ),
    dict(
        name="logpar",
        model=LogParabolaSpectralModel(
            alpha=2.3 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            beta=0.5 * u.Unit(""),
        ),
        val_at_2TeV=u.Quantity(0.6387956571420305, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.255689748270628, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(3.9586515834989267, "TeV cm-2 s-1"),
        e_peak=0.74082 * u.TeV,
    ),
    dict(
        name="norm-logpar",
        model=LogParabolaNormSpectralModel(
            alpha=2.3 * u.Unit(""),
            norm=4 * u.Unit(""),
            reference=1 * u.TeV,
            beta=0.5 * u.Unit(""),
        ),
        val_at_2TeV=u.Quantity(0.6387956571420305, ""),
        integral_1_10TeV=u.Quantity(2.255689748270628, "TeV"),
        eflux_1_10TeV=u.Quantity(3.9586515834989267, "TeV2"),
    ),
    dict(
        name="logpar10",
        model=LogParabolaSpectralModel.from_log10(
            alpha=2.3 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            beta=1.151292546497023 * u.Unit(""),
        ),
        val_at_2TeV=u.Quantity(0.6387956571420305, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.255689748270628, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(3.9586515834989267, "TeV cm-2 s-1"),
        e_peak=0.74082 * u.TeV,
    ),
    dict(
        name="powerlaw_index1",
        model=PowerLawSpectralModel(
            index=1 * u.Unit(""),
            amplitude=2 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(1.0, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(4.605170185, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(18.0, "TeV cm-2 s-1"),
    ),
    dict(
        name="ecpl_2",
        model=ExpCutoffPowerLawSpectralModel(
            index=2.0 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            lambda_=0.1 / u.TeV,
        ),
        val_at_2TeV=u.Quantity(0.81873075, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.83075297, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(6.41406327, "TeV cm-2 s-1"),
        e_peak=np.nan * u.TeV,
    ),
    dict(
        name="GaussianSpectralModel",
        model=GaussianSpectralModel(
            amplitude=4 / u.cm**2 / u.s, mean=2 * u.TeV, sigma=0.2 * u.TeV
        ),
        val_at_2TeV=u.Quantity(7.978845608028654, "cm-2 s-1 TeV-1"),
        val_at_3TeV=u.Quantity(2.973439029468601e-05, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(3.9999988533937123, "cm-2 s-1"),
        integral_infinity=u.Quantity(4, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(7.999998896163037, "TeV cm-2 s-1"),
    ),
    dict(
        name="ecpl",
        model=ExpCutoffPowerLawSpectralModel(
            index=1.8 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            reference=1 * u.TeV,
            lambda_=0.1 / u.TeV,
            alpha=0.8,
        ),
        val_at_2TeV=u.Quantity(0.871694294554192, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(3.026342, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(7.38652453, "TeV cm-2 s-1"),
        e_peak=1.7677669529663684 * u.TeV,
    ),
    dict(
        name="bpl",
        model=BrokenPowerLawSpectralModel(
            index1=1.5 * u.Unit(""),
            index2=2.5 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            ebreak=0.5 * u.TeV,
        ),
        val_at_2TeV=u.Quantity(0.125, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(0.45649740094103286, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(0.9669999668731384, "TeV cm-2 s-1"),
    ),
    dict(
        name="sbpl",
        model=SmoothBrokenPowerLawSpectralModel(
            index1=1.5 * u.Unit(""),
            index2=2.5 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            ebreak=0.5 * u.TeV,
            reference=1 * u.TeV,
            beta=1,
        ),
        val_at_2TeV=u.Quantity(0.28284271247461906, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(0.9956923907948155, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(2.2372256145972207, "TeV cm-2 s-1"),
    ),
    dict(
        name="sbpl-hard",
        model=SmoothBrokenPowerLawSpectralModel(
            index1=2.5 * u.Unit(""),
            index2=1.5 * u.Unit(""),
            amplitude=4 / u.cm**2 / u.s / u.TeV,
            ebreak=0.5 * u.TeV,
            reference=1 * u.TeV,
            beta=1,
        ),
        val_at_2TeV=u.Quantity(3.5355339059327378, "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(13.522782989735022, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(40.06681812966845, "TeV cm-2 s-1"),
    ),
    dict(
        name="pbpl",
        model=PiecewiseNormSpectralModel(
            energy=[1, 3, 7, 10] * u.TeV,
            norms=[1, 5, 3, 0.5] * u.Unit(""),
        ),
        val_at_2TeV=u.Quantity(2.76058404, ""),
        integral_1_10TeV=u.Quantity(24.758255, "TeV"),
        eflux_1_10TeV=u.Quantity(117.745068, "TeV2"),
    ),
    dict(
        name="pbpllin",
        model=PiecewiseNormSpectralModel(
            energy=[1, 3, 7, 10] * u.TeV,
            norms=[1, 5, 3, 0.5] * u.Unit(""),
            interp="lin",
        ),
        val_at_2TeV=u.Quantity(3.0, ""),
        integral_1_10TeV=u.Quantity(27.24066846, "TeV"),
        eflux_1_10TeV=u.Quantity(133.34487, "TeV2"),
    ),
]

# Add compound models

TEST_MODELS.append(
    dict(
        name="compound6",
        model=TEST_MODELS[0]["model"] + u.Quantity(4, "cm-2 s-1 TeV-1"),
        val_at_2TeV=TEST_MODELS[0]["val_at_2TeV"] * 2,
        integral_1_10TeV=TEST_MODELS[0]["integral_1_10TeV"] * 2,
        eflux_1_10TeV=TEST_MODELS[0]["eflux_1_10TeV"] * 2,
    )
)

TEST_MODELS.append(
    dict(
        name="compound3",
        model=TEST_MODELS[1]["model"] + TEST_MODELS[1]["model"],
        val_at_2TeV=TEST_MODELS[1]["val_at_2TeV"] * 2,
        integral_1_10TeV=TEST_MODELS[1]["integral_1_10TeV"] * 2,
        eflux_1_10TeV=TEST_MODELS[1]["eflux_1_10TeV"] * 2,
    )
)


TEST_MODELS.append(
    dict(
        name="table_model",
        model=table_model(),
        # Values took from power law expectation
        val_at_2TeV=u.Quantity(4 * 2.0 ** (-2.3), "cm-2 s-1 TeV-1"),
        integral_1_10TeV=u.Quantity(2.9227116204223784, "cm-2 s-1"),
        eflux_1_10TeV=u.Quantity(6.650836884969039, "TeV cm-2 s-1"),
    )
)


@requires_dependency("scipy")
@pytest.mark.parametrize("spectrum", TEST_MODELS, ids=lambda _: _["name"])
def test_models(spectrum):
    model = spectrum["model"]

    for p in model.parameters:
        assert p.type == "spectral"

    energy = 2 * u.TeV
    value = model(energy)
    energies = [2, 3] * u.TeV
    values = model(energies)
    assert_quantity_allclose(value, spectrum["val_at_2TeV"], rtol=1e-7)
    if "val_at_3TeV" in spectrum:
        energy = 3 * u.TeV
        value = model(energy)
        assert_quantity_allclose(value, spectrum["val_at_3TeV"], rtol=1e-7)

    energy_min = 1 * u.TeV
    energy_max = 10 * u.TeV
    assert_quantity_allclose(
        model.integral(energy_min=energy_min, energy_max=energy_max),
        spectrum["integral_1_10TeV"],
        rtol=1e-5,
    )
    assert_quantity_allclose(
        model.energy_flux(energy_min=energy_min, energy_max=energy_max),
        spectrum["eflux_1_10TeV"],
        rtol=1e-5,
    )

    if "e_peak" in spectrum:
        assert_quantity_allclose(model.e_peak, spectrum["e_peak"], rtol=1e-2)

    # inverse for ConstantSpectralModel is irrelevant.
    # inverse for Gaussian and PiecewiseNormSpectralModel have a degeneracy
    if not (
        isinstance(model, ConstantSpectralModel)
        or spectrum["name"] == "compound6"
        or spectrum["name"] == "GaussianSpectralModel"
        or "pbpl" in spectrum["name"]
    ):
        assert_quantity_allclose(model.inverse(value), energy, rtol=0.01)
        inverse = model.inverse_all(values)
        for ke, ener in enumerate(energies):
            assert_quantity_allclose(inverse[ke], energies[ke], rtol=0.01)

    if "integral_infinity" in spectrum:
        energy_min = 0 * u.TeV
        energy_max = 10000 * u.TeV
        assert_quantity_allclose(
            model.integral(energy_min=energy_min, energy_max=energy_max),
            spectrum["integral_infinity"],
        )

    model.to_dict()

    assert "" in str(model)

    # check that an array evaluation works (otherwise e.g. plotting raises an error)
    e_array = [2, 10, 20] * u.TeV
    e_array = e_array[:, np.newaxis, np.newaxis]
    val = model(e_array)
    assert val.shape == e_array.shape
    assert_quantity_allclose(val[0], spectrum["val_at_2TeV"])


def test_evaluate():
    for m in TEST_MODELS:
        model = m["model"]
        energies = [1e-12, 1e-6, 1e-2, 1, 1e2, 1e4] * u.TeV
        parameters = model.parameters
        par_list = [p.quantity for p in parameters]
        result_eval = model.evaluate(energies, *par_list)

        result_call = model(energies)
        assert_quantity_allclose(result_eval, result_call)


def test_model_unit():
    pwl = PowerLawSpectralModel()
    value = pwl(500 * u.MeV)
    assert value.unit == "cm-2 s-1 TeV-1"


def test_model_plot():
    pwl = PowerLawSpectralModel(
        amplitude=1e-12 * u.Unit("TeV-1 cm-2 s-1"), reference=1 * u.Unit("TeV"), index=2
    )
    pwl.amplitude.error = 0.1e-12 * u.Unit("TeV-1 cm-2 s-1")
    energy_axis = MapAxis.from_energy_bounds(1 * u.TeV, 10 * u.TeV, 100)

    with mpl_plot_check():
        pwl.plot((1 * u.TeV, 10 * u.TeV))

    with mpl_plot_check():
        pwl.plot_error((1 * u.TeV, 10 * u.TeV))

    with mpl_plot_check():
        pwl.plot([0.08, 20] * u.TeV)

    with mpl_plot_check():
        pwl.plot_error([0.08, 20] * u.TeV)

    with mpl_plot_check():
        pwl.plot([0.08 * u.TeV, 20 * u.TeV])

    with mpl_plot_check():
        pwl.plot_error([0.08 * u.TeV, 20 * u.TeV])

    with mpl_plot_check():
        pwl.plot(energy_bounds=energy_axis)

    with mpl_plot_check():
        pwl.plot_error(energy_bounds=energy_axis)


def test_model_plot_sed_type():
    pwl = PowerLawSpectralModel(
        amplitude=1e-12 * u.Unit("TeV-1 cm-2 s-1"), reference=1 * u.Unit("TeV"), index=2
    )
    pwl.amplitude.error = 0.1e-12 * u.Unit("TeV-1 cm-2 s-1")

    with mpl_plot_check():
        ax1 = pwl.plot((1 * u.TeV, 100 * u.TeV), sed_type="dnde")
        ax2 = pwl.plot_error((1 * u.TeV, 100 * u.TeV), sed_type="dnde")
        assert ax1.yaxis.units == u.Unit("1 / (s cm2 TeV)")
        assert ax1.axes.axes.get_ylabel().split()[0] == "dnde"
        assert ax2.yaxis.units == u.Unit("1 / (s cm2 TeV)")
        assert ax2.axes.axes.get_ylabel().split()[0] == "dnde"

    with mpl_plot_check():
        ax1 = pwl.plot((1 * u.TeV, 100 * u.TeV), sed_type="e2dnde")
        ax2 = pwl.plot_error((1 * u.TeV, 100 * u.TeV), sed_type="e2dnde")
        assert ax1.yaxis.units == u.Unit("erg / (cm2 s)")
        assert ax1.axes.axes.get_ylabel().split()[0] == "e2dnde"
        assert ax2.yaxis.units == u.Unit("erg / (cm2 s)")
        assert ax2.axes.axes.get_ylabel().split()[0] == "e2dnde"

    with mpl_plot_check():
        ax1 = pwl.plot((1 * u.TeV, 100 * u.TeV), sed_type="flux")
        ax2 = pwl.plot_error((1 * u.TeV, 100 * u.TeV), sed_type="flux")
        assert ax1.yaxis.units == u.Unit("1 / (s cm2)")
        assert ax1.axes.axes.get_ylabel().split()[0] == "flux"
        assert ax2.yaxis.units == u.Unit("1 / (s cm2)")
        assert ax2.axes.axes.get_ylabel().split()[0] == "flux"

    with mpl_plot_check():
        ax1 = pwl.plot((1 * u.TeV, 100 * u.TeV), sed_type="eflux")
        ax2 = pwl.plot_error((1 * u.TeV, 100 * u.TeV), sed_type="eflux")
        assert ax1.yaxis.units == u.Unit("erg / (cm2 s)")
        assert ax1.axes.axes.get_ylabel().split()[0] == "eflux"
        assert ax2.yaxis.units == u.Unit("erg / (cm2 s)")
        assert ax2.axes.axes.get_ylabel().split()[0] == "eflux"


def test_to_from_dict():
    spectrum = TEST_MODELS[1]
    model = spectrum["model"]

    model_dict = model.to_dict()
    # Here we reverse the order of parameters list to ensure assignment is correct
    model_dict["spectral"]["parameters"].reverse()

    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)

    assert isinstance(new_model, PowerLawSpectralModel)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)

    actual = [par.frozen for par in new_model.parameters]
    desired = [par.frozen for par in model.parameters]
    assert_allclose(actual, desired)

    new_model = Model.from_dict(model_dict)

    assert isinstance(new_model, PowerLawSpectralModel)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)

    actual = [par.frozen for par in new_model.parameters]
    desired = [par.frozen for par in model.parameters]
    assert_allclose(actual, desired)


def test_to_from_dict_partial_input(caplog):
    spectrum = TEST_MODELS[1]
    model = spectrum["model"]

    model_dict = model.to_dict()
    # Here we remove the reference energy
    model_dict["spectral"]["parameters"].remove(model_dict["spectral"]["parameters"][2])

    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)

    assert isinstance(new_model, PowerLawSpectralModel)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)

    actual = [par.frozen for par in new_model.parameters]
    desired = [par.frozen for par in model.parameters]
    assert_allclose(actual, desired)
    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert (
        "Parameter 'reference' not defined in YAML file. Using default value: 1.0 TeV"
        in [_.message for _ in caplog.records]
    )


def test_to_from_dict_compound():
    spectrum = TEST_MODELS[-3]
    model = spectrum["model"]
    assert spectrum["name"] == "compound6"
    model_dict = model.to_dict()
    assert model_dict["spectral"]["operator"] == "add"
    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)

    assert isinstance(new_model, CompoundSpectralModel)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)


def test_to_from_dict_piecewise_lin():
    spectrum = TEST_MODELS[-4]
    model = spectrum["model"]
    assert spectrum["name"] == "pbpllin"
    model_dict = model.to_dict()
    assert model_dict["spectral"]["interp"] == "lin"
    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)
    assert isinstance(new_model, PiecewiseNormSpectralModel)
    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)
    assert new_model._interp == model._interp

    model_dict["spectral"].pop("interp")
    new_model_default = model_class.from_dict(model_dict)
    assert isinstance(new_model_default, PiecewiseNormSpectralModel)
    assert new_model_default._interp == "log"


def test_to_from_dict_piecewise():
    spectrum = TEST_MODELS[-5]
    model = spectrum["model"]
    assert spectrum["name"] == "pbpl"
    model_dict = model.to_dict()
    assert model_dict["spectral"]["interp"] == "log"
    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)
    assert isinstance(new_model, PiecewiseNormSpectralModel)
    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)
    assert new_model._interp == model._interp

    model_dict["spectral"].pop("interp")
    new_model_default = model_class.from_dict(model_dict)
    assert isinstance(new_model_default, PiecewiseNormSpectralModel)
    assert new_model_default._interp == "log"


@requires_data()
def test_table_model_from_file():
    filename = "$GAMMAPY_DATA/ebl/ebl_franceschini.fits.gz"
    absorption_z03 = TemplateSpectralModel.read_xspec_model(
        filename=filename, param=0.3
    )
    value = absorption_z03(1 * u.TeV)
    assert_allclose(value, 1)


@requires_data()
def test_absorption():
    # absorption values for given redshift
    redshift = 0.117
    absorption = EBLAbsorptionNormSpectralModel.read_builtin(
        "dominguez", redshift=redshift
    )

    # Spectral model corresponding to PKS 2155-304 (quiescent state)
    index = 3.53
    amplitude = 1.81 * 1e-12 * u.Unit("cm-2 s-1 TeV-1")
    reference = 1 * u.TeV
    pwl = PowerLawSpectralModel(index=index, amplitude=amplitude, reference=reference)

    # Test error propagation
    pwl.amplitude.error = 0.1 * pwl.amplitude.value
    dnde, dnde_errn, dnde_errp = pwl.evaluate_error(1 * u.TeV, n_samples=20000)
    assert_allclose((dnde_errn / dnde), 0.1, rtol=7e-1)
    assert_allclose((dnde_errp / dnde), 0.1, rtol=7e-1)

    # EBL + PWL model
    model = pwl * absorption
    desired = u.Quantity(5.140765e-13, "TeV-1 s-1 cm-2")
    assert_quantity_allclose(model(1 * u.TeV), desired, rtol=1e-3)
    assert model.model2.alpha_norm.value == 1.0

    # EBL + PWL model: test if norm of EBL=0: it mean model =pwl
    model.parameters["alpha_norm"].value = 0
    assert_quantity_allclose(model(1 * u.TeV), pwl(1 * u.TeV), rtol=1e-3)

    # EBL + PWL model: Test with a norm different of 1
    absorption = EBLAbsorptionNormSpectralModel.read_builtin(
        "dominguez", redshift=redshift, alpha_norm=1.5
    )
    model = pwl * absorption
    desired = u.Quantity(2.739695e-13, "TeV-1 s-1 cm-2")
    assert model.model2.alpha_norm.value == 1.5
    assert_quantity_allclose(model(1 * u.TeV), desired, rtol=1e-3)

    # Test error propagation
    dnde, dnde_errn, dnde_errp = model.evaluate_error(1 * u.TeV, n_samples=20000)
    assert_allclose((dnde_errn / dnde).data, 0.1, rtol=3e-2)
    assert_allclose((dnde_errp / dnde).data, 0.1, rtol=3e-2)


@requires_data()
def test_absorbed_extrapolate():
    ebl_model = "dominguez"
    z = 0.0001
    alpha_norm = 1
    absorption = EBLAbsorptionNormSpectralModel.read_builtin(ebl_model)

    values = absorption.evaluate(1 * u.TeV, z, alpha_norm)
    assert_allclose(values, 1)


def test_ecpl_integrate():
    # regression test to check the numerical integration for small energy bins
    ecpl = ExpCutoffPowerLawSpectralModel()
    value = ecpl.integral(1 * u.TeV, 1.1 * u.TeV)
    assert value.isscalar
    assert_quantity_allclose(value, 8.380714e-14 * u.Unit("s-1 cm-2"))


def test_call_plsec_4fgl_dr1():
    model = SuperExpCutoffPowerLaw4FGLSpectralModel(
        amplitude="2e-12 MeV-1 s-1 cm-2",
        reference="1000 MeV",
        expfactor=5e-03,
        index_1=2,
        index_2=1,
    )
    desired = u.Quantity(3.823779e-20, "MeV-1 s-1 cm-2")
    assert_allclose(model(4 * u.GeV), desired, rtol=1e-3)


def test_pwl_pivot_energy():
    pwl = PowerLawSpectralModel(amplitude="5.35510540e-11 cm-2 s-1 TeV-1")
    assert_quantity_allclose(pwl.pivot_energy, np.nan * u.TeV, rtol=1e-5)

    pwl.covariance = [
        [0.08**2, 6.56889e-14, 0],
        [6.56889e-14, (5.5e-12) ** 2, 0],
        [0, 0, 0],
    ]
    assert_quantity_allclose(pwl.pivot_energy, 1.211265 * u.TeV, rtol=1e-5)

    ecpl = ExpCutoffPowerLawSpectralModel(
        amplitude="5.35510540e-11 cm-2 s-1 TeV-1", lambda_=0.001 * (1 / u.TeV), index=2
    )
    ecpl.covariance = [
        [0.08**2, 6.56889e-14, 0, 0, 0],
        [6.56889e-14, (5.5e-12) ** 2, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
    ]
    assert_quantity_allclose(ecpl.pivot_energy, 1.263819 * u.TeV, rtol=5e-2)


def test_num_pivot_energy():
    lp = LogParabolaSpectralModel(
        amplitude="5.82442e-11 cm-2 s-1 GeV-1",
        reference="17.337 GeV",
        alpha="1.9134",
        beta="0.2217",
    )
    lp.amplitude.error = "2.8804e-12 cm-2 s-1 GeV-1"
    assert_quantity_allclose(lp.pivot_energy, np.nan * u.GeV, rtol=1e-5)

    lp.alpha.error = "0.1126"
    lp.beta.error = "0.0670"
    assert_quantity_allclose(lp.pivot_energy, 18.198706 * u.GeV, rtol=7e-2)


def test_template_spectral_model_evaluate_tiny():
    energy = np.array([1.00000000e06, 1.25892541e06, 1.58489319e06, 1.99526231e06])
    values = np.array([4.39150790e-38, 1.96639562e-38, 8.80497507e-39, 3.94262401e-39])

    model = TemplateSpectralModel(
        energy=energy, values=values * u.Unit("MeV-1 s-1 sr-1")
    )
    result = model(energy)
    tiny = np.finfo(np.float32).tiny
    mask = abs(values) - tiny > tiny
    np.testing.assert_allclose(
        values[mask] / values.max(), result[mask].value / values.max()
    )
    mask = abs(result.value) - tiny <= tiny
    assert np.all(result[mask] == 0.0)


def test_template_spectral_model_single_value():
    energy = [1] * u.TeV
    values = [1e-12] * u.Unit("TeV-1 s-1 cm-2")

    model = TemplateSpectralModel(energy=energy, values=values)
    result = model(energy=[0.5, 2] * u.TeV)

    assert_allclose(result.data, 1e-12)

    data = model.to_dict()
    model2 = TemplateSpectralModel.from_dict(data)
    assert model2.to_dict() == data


def test_template_spectral_model_compound():
    energy = [1.00e06, 1.25e06, 1.58e06, 1.99e06] * u.MeV
    values = [4.39e-7, 1.96e-7, 8.80e-7, 3.94e-7] * u.Unit("MeV-1 s-1 sr-1")

    template = TemplateSpectralModel(energy=energy, values=values)
    correction = PowerLawNormSpectralModel(norm=2)
    model = CompoundSpectralModel(template, correction, operator=operator.mul)
    assert np.allclose(model(energy), 2 * values)

    model_mul = template * correction
    assert isinstance(model_mul, CompoundSpectralModel)
    assert np.allclose(model_mul(energy), 2 * values)

    model_dict = model.to_dict()
    assert model_dict["spectral"]["operator"] == "mul"
    model_class = SPECTRAL_MODEL_REGISTRY.get_cls(model_dict["spectral"]["type"])
    new_model = model_class.from_dict(model_dict)
    assert isinstance(new_model, CompoundSpectralModel)
    assert np.allclose(new_model(energy), 2 * values)


def test_template_spectral_model_options():
    energy = [1.00e06, 1.25e06, 1.58e06, 1.99e06] * u.MeV
    values = [4.39e-7, 1.96e-7, 8.80e-7, 3.94e-7] * u.Unit("MeV-1 s-1 sr-1")

    model = TemplateSpectralModel(
        energy=energy,
        values=values,
        interp_kwargs={"extrapolate": True, "method": "linear"},
    )
    assert np.allclose(model(energy), values)


def test_covariance_spectral_model_compound():
    model = TEST_MODELS[2]["model"] + TEST_MODELS[1]["model"]
    covar = np.eye(len(model.parameters))
    covar[0, -1] = 1.0
    covar[-1, 0] = 1.0
    covar[0, 1] = 1.0
    covar[-1, -3] = 1.0

    model.covariance = covar
    assert_allclose(model.covariance.data, covar)
    assert_allclose(model.model1.covariance.data, covar[:3, :3])
    assert_allclose(model.model2.covariance.data, covar[-3:, -3:])


def test_evaluate_spectral_model_compound():
    energy = [1.00e06, 1.25e06, 1.58e06, 1.99e06] * u.MeV
    model = TEST_MODELS[-2]["model"]
    values = model.evaluate(energy, *[p.quantity for p in model.parameters])
    assert_allclose(values, model(energy))

    model = TEST_MODELS[-3]["model"]
    values = model.evaluate(energy, *[p.quantity for p in model.parameters])
    assert_allclose(values, model(energy))


def test_evaluate_nested_spectral_model_compound():
    energy = [1.00e06, 1.25e06, 1.58e06, 1.99e06] * u.MeV
    model1 = TEST_MODELS[-2]["model"]
    model2 = TEST_MODELS[-3]["model"]

    model = model1 + model2
    values = model.evaluate(energy, *[p.quantity for p in model.parameters])
    assert_allclose(values, model(energy))


@requires_dependency("naima")
class TestNaimaModel:
    # Used to test model value at 2 TeV
    energy = 2 * u.TeV

    # Used to test model integral and energy flux
    energy_min = 1 * u.TeV
    energy_max = 10 * u.TeV

    # Used to that if array evaluation works
    e_array = [2, 10, 20] * u.TeV
    e_array = e_array[:, np.newaxis, np.newaxis]

    def test_pion_decay(self):
        import naima

        particle_distribution = naima.models.PowerLaw(
            amplitude=2e33 / u.eV, e_0=10 * u.TeV, alpha=2.5
        )
        radiative_model = naima.radiative.PionDecay(
            particle_distribution, nh=1 * u.cm**-3
        )
        model = NaimaSpectralModel(radiative_model)
        for p in model.parameters:
            assert p._type == "spectral"

        val_at_2TeV = 9.725347355450884e-14 * u.Unit("cm-2 s-1 TeV-1")
        integral_1_10TeV = 3.530537143620737e-13 * u.Unit("cm-2 s-1")
        eflux_1_10TeV = 7.643559573105779e-13 * u.Unit("TeV cm-2 s-1")

        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV)
        assert_quantity_allclose(
            model.integral(energy_min=self.energy_min, energy_max=self.energy_max),
            integral_1_10TeV,
        )
        assert_quantity_allclose(
            model.energy_flux(energy_min=self.energy_min, energy_max=self.energy_max),
            eflux_1_10TeV,
        )
        val = model(self.e_array)
        assert val.shape == self.e_array.shape

        model.amplitude.error = 0.1 * model.amplitude.value

        out = model.evaluate_error(1 * u.TeV)
        assert_allclose(out.data, [5.270367e-13, 5.293272e-14, 5.343376e-14], rtol=5e-2)

    def test_ic(self):
        import naima

        particle_distribution = naima.models.ExponentialCutoffBrokenPowerLaw(
            amplitude=2e33 / u.eV,
            e_0=10 * u.TeV,
            alpha_1=2.5,
            alpha_2=2.7,
            e_break=900 * u.GeV,
            e_cutoff=10 * u.TeV,
        )
        radiative_model = naima.radiative.InverseCompton(
            particle_distribution, seed_photon_fields=["CMB"]
        )

        model = NaimaSpectralModel(radiative_model)
        for p in model.parameters:
            assert p._type == "spectral"

        val_at_2TeV = 4.347836316893546e-12 * u.Unit("cm-2 s-1 TeV-1")
        integral_1_10TeV = 1.595813e-11 * u.Unit("cm-2 s-1")
        eflux_1_10TeV = 2.851283e-11 * u.Unit("TeV cm-2 s-1")

        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV)
        assert_quantity_allclose(
            model.integral(energy_min=self.energy_min, energy_max=self.energy_max),
            integral_1_10TeV,
            rtol=1e-5,
        )
        assert_quantity_allclose(
            model.energy_flux(energy_min=self.energy_min, energy_max=self.energy_max),
            eflux_1_10TeV,
            rtol=1e-5,
        )
        val = model(self.e_array)
        assert val.shape == self.e_array.shape

    def test_synchrotron(self):
        import naima

        particle_distribution = naima.models.LogParabola(
            amplitude=2e33 / u.eV, e_0=10 * u.TeV, alpha=1.3, beta=0.5
        )
        radiative_model = naima.radiative.Synchrotron(particle_distribution, B=2 * u.G)

        model = NaimaSpectralModel(radiative_model)
        for p in model.parameters:
            assert p._type == "spectral"

        val_at_2TeV = 1.0565840392550432e-24 * u.Unit("cm-2 s-1 TeV-1")
        integral_1_10TeV = 4.449186e-13 * u.Unit("cm-2 s-1")
        eflux_1_10TeV = 4.594121e-13 * u.Unit("TeV cm-2 s-1")

        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV)
        assert_quantity_allclose(
            model.integral(energy_min=self.energy_min, energy_max=self.energy_max),
            integral_1_10TeV,
            rtol=1e-5,
        )
        assert_quantity_allclose(
            model.energy_flux(energy_min=self.energy_min, energy_max=self.energy_max),
            eflux_1_10TeV,
            rtol=1e-5,
        )
        val = model(self.e_array)
        assert val.shape == self.e_array.shape

        model.B.value = 3  # update B
        val_at_2TeV = 5.1985064062296e-16 * u.Unit("cm-2 s-1 TeV-1")
        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV)

    def test_ssc(self):
        import naima

        ECBPL = naima.models.ExponentialCutoffBrokenPowerLaw(
            amplitude=3.699e36 / u.eV,
            e_0=1 * u.TeV,
            e_break=0.265 * u.TeV,
            alpha_1=1.5,
            alpha_2=3.233,
            e_cutoff=1863 * u.TeV,
            beta=2.0,
        )

        radiative_model = naima.radiative.InverseCompton(
            ECBPL,
            seed_photon_fields=[
                "CMB",
                ["FIR", 70 * u.K, 0.5 * u.eV / u.cm**3],
                ["NIR", 5000 * u.K, 1 * u.eV / u.cm**3],
            ],
            Eemax=50 * u.PeV,
            Eemin=0.1 * u.GeV,
        )
        B = 125 * u.uG
        radius = 2.1 * u.pc
        nested_models = {"SSC": {"B": B, "radius": radius}}
        model = NaimaSpectralModel(radiative_model, nested_models=nested_models)
        assert_quantity_allclose(model.B.quantity, B)
        assert_quantity_allclose(model.radius.quantity, radius)
        val_at_2TeV = 1.6703761561806372e-11 * u.Unit("cm-2 s-1 TeV-1")
        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV, rtol=1e-5)

        model.parameters["B"].value = 100
        val_at_2TeV = 1.441331153167876e-11 * u.Unit("cm-2 s-1 TeV-1")
        value = model(self.energy)
        assert_quantity_allclose(value, val_at_2TeV, rtol=1e-5)

    def test_bad_init(self):
        import naima

        particle_distribution = naima.models.PowerLaw(
            amplitude=2e33 / u.eV, e_0=10 * u.TeV, alpha=2.5
        )
        radiative_model = naima.radiative.PionDecay(
            particle_distribution, nh=1 * u.cm**-3
        )
        model = NaimaSpectralModel(radiative_model)

        with pytest.raises(NotImplementedError):
            NaimaSpectralModel.from_dict(model.to_dict())
        with pytest.raises(NotImplementedError):
            NaimaSpectralModel.from_parameters(model.parameters)

    def test_cache(self):
        import naima

        particle_distribution = naima.models.ExponentialCutoffPowerLaw(
            1e30 / u.eV, 10 * u.TeV, 3.0, 30 * u.TeV
        )
        radiative_model = naima.radiative.InverseCompton(
            particle_distribution,
            seed_photon_fields=["CMB", ["FIR", 26.5 * u.K, 0.415 * u.eV / u.cm**3]],
            Eemin=100 * u.GeV,
        )
        model = NaimaSpectralModel(radiative_model, distance=1.5 * u.kpc)

        opts = {
            "energy_bounds": [10 * u.GeV, 80 * u.TeV],
            "sed_type": "e2dnde",
        }
        _, ax = plt.subplots()
        model.plot(label="IC (total)", ax=ax, **opts)

        for seed, ls in zip(["CMB", "FIR"], ["-", "--"]):
            model = NaimaSpectralModel(radiative_model, seed=seed, distance=1.5 * u.kpc)
            model.plot(label=f"IC ({seed})", ls=ls, color="gray", **opts)

        skymodel = SkyModel(model)
        # fail if cache is on :
        _, ax = plt.subplots()
        skymodel.spectral_model.plot(
            energy_bounds=[10 * u.GeV, 80 * u.TeV], energy_power=2, ax=ax
        )
        assert not radiative_model._memoize


class TestSpectralModelErrorPropagation:
    """Test spectral model error propagation.

    https://github.com/gammapy/gammapy/blob/master/docs/development/pigs/pig-014.rst#proposal
    https://nbviewer.jupyter.org/github/gammapy/gammapy-extra/blob/master/experiments/uncertainty_estimation_prototype.ipynb
    """

    def setup_method(self):
        self.model = LogParabolaSpectralModel(
            amplitude=3.76e-11 * u.Unit("cm-2 s-1 TeV-1"),
            reference=1 * u.TeV,
            alpha=2.44,
            beta=0.25,
        )
        self.model.covariance = [
            [1.31e-23, 0, -6.80e-14, 3.04e-13],
            [0, 0, 0, 0],
            [-6.80e-14, 0, 0.00899, 0.00904],
            [3.04e-13, 0, 0.00904, 0.0284],
        ]

    def test_evaluate_error_scalar(self):
        # evaluate_error on scalar
        out = self.model.evaluate_error(1 * u.TeV)
        assert isinstance(out, u.Quantity)
        assert out.unit == "cm-2 s-1 TeV-1"
        assert out.shape == (3,)
        assert_allclose(out.data, [3.757824e-11, 3.142095e-12, 3.155082e-12], rtol=7e-2)

    def test_evaluate_error_array(self):
        out = self.model.evaluate_error([1, 100] * u.TeV)
        assert out.shape == (3, 2)
        expected = [
            [3.756851e-11, 2.600163e-18],
            [3.147163e-12, 2.539555e-18],
            [3.163654e-12, 1.082997e-16],
        ]
        assert_allclose(out.data, expected, rtol=7e-1)

    def test_evaluate_error_unit(self):
        out = self.model.evaluate_error(1e6 * u.MeV)
        assert out.unit == "cm-2 s-1 TeV-1"
        assert_allclose(out.data, [3.757824e-11, 3.142095e-12, 3.155082e-12], rtol=7e-2)

        with pytest.warns(GammapyDeprecationWarning):
            self.model.evaluate_error(1e6 * u.MeV, epsilon=10)

    def test_integral_error(self):
        out = self.model.integral_error(1 * u.TeV, 10 * u.TeV)
        assert out.unit == "cm-2 s-1"
        assert out.shape == (3,)
        assert_allclose(out.data, [2.205520e-11, 2.395797e-12, 2.905355e-12], rtol=7e-2)

        with pytest.warns(GammapyDeprecationWarning):
            self.model.integral_error(1 * u.TeV, 10 * u.TeV, epsilon=10)

    def test_energy_flux_error(self):
        out = self.model.energy_flux_error(1 * u.TeV, 10 * u.TeV)
        assert out.unit == "TeV cm-2 s-1"
        assert out.shape == (3,)
        assert_allclose(out.data, [4.133955e-11, 6.634629e-12, 9.896814e-12], rtol=7e-2)

        with pytest.warns(GammapyDeprecationWarning):
            self.model.energy_flux_error(1 * u.TeV, 10 * u.TeV, epsilon=10)


def test_piecesenorm_model_error():
    model = PiecewiseNormSpectralModel(
        energy=[1, 3, 7, 10] * u.TeV,
        norms=[1, 5, 3, 0.5] * u.Unit(""),
    )
    for p in model.parameters:
        p.error = 1.0
    assert_allclose(model.evaluate_error(5 * u.GeV), 1.0, rtol=5e-2)
    assert_allclose(model.integral_error(5 * u.GeV, 10 * u.GeV).value, 5.0, rtol=5e-2)


def test_logpar_index_error():
    model = LogParabolaSpectralModel(
        amplitude=3.81e-11 / u.cm**2 / u.s / u.TeV,
        reference=1 * u.TeV,
        alpha=2.19,
        beta=0.226,
    )
    model.alpha.error = 0.4
    out = model.spectral_index_error(energy=1.0 * u.TeV)
    assert_allclose(out, [2.198321, 0.404178, 0.394804], rtol=5e-2)


def test_dnde_error_ecpl_model():
    # Regression test for ECPL model
    # https://github.com/gammapy/gammapy/issues/2007
    model = ExpCutoffPowerLawSpectralModel(
        amplitude=2.076183759227292e-12 * u.Unit("cm-2 s-1 TeV-1"),
        index=1.8763343736076483,
        lambda_=0.08703226432146616 * u.Unit("TeV-1"),
        reference=1 * u.TeV,
    )
    model.covariance = [
        [0.00204191498, -1.507724e-14, 0.0, -0.001834819, 0.0],
        [-1.507724e-14, 1.6864740e-25, 0.0, 1.854251e-14, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0],
        [-0.001834819175, 1.8542517e-14, 0.0, 0.0032559101, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0],
    ]

    out = model.evaluate_error(1 * u.TeV)
    assert_allclose(out.data, [1.900058e-12, 3.140516e-13, 2.925388e-13], rtol=7e-1)

    out = model.evaluate_error(0.1 * u.TeV)
    assert_allclose(out.data, [1.538462e-10, 2.071542e-11, 1.837818e-11], rtol=7e-1)

    with pytest.warns(GammapyDeprecationWarning):
        out = model.evaluate_error(0.1 * u.TeV, epsilon=10)


def test_integral_error_power_law():
    energy = np.linspace(1 * u.TeV, 10 * u.TeV, 10)
    energy_min = energy[:-1]
    energy_max = energy[1:]

    powerlaw = PowerLawSpectralModel()
    powerlaw.parameters["index"].error = 0.4
    powerlaw.parameters["amplitude"].error = 1e-13

    flux, flux_errn, flux_errp = powerlaw.integral_error(energy_min, energy_max)

    assert_allclose(flux.value[0] / 1e-13, 4.991835, rtol=1e-3)
    assert_allclose(flux_errn.value[0] / 1e-14, 7.209218, rtol=7e-1)
    assert_allclose(flux_errp.value[0] / 1e-14, 8.674678, rtol=7e-1)


def test_integral_error_exp_cut_off_power_law():
    energy = np.linspace(1 * u.TeV, 10 * u.TeV, 10)
    energy_min = energy[:-1]
    energy_max = energy[1:]

    exppowerlaw = ExpCutoffPowerLawSpectralModel()
    exppowerlaw.parameters["index"].error = 0.4
    exppowerlaw.parameters["amplitude"].error = 1e-13
    exppowerlaw.parameters["lambda_"].error = 0.03

    flux, flux_errn, flux_errp = exppowerlaw.integral_error(energy_min, energy_max)

    assert_allclose(flux.value[0] / 1e-13, 5.058788, rtol=0.01)
    assert_allclose(flux_errn.value[0] / 1e-14, 7.79546, rtol=7e-1)
    assert_allclose(flux_errp.value[0] / 1e-14, 9.537771, rtol=7e-1)


def test_energy_flux_error_power_law():
    energy_min = 1 * u.TeV
    energy_max = 10 * u.TeV

    powerlaw = PowerLawSpectralModel()
    powerlaw.parameters["index"].error = 0.4
    powerlaw.parameters["amplitude"].error = 1e-13

    eflux, eflux_errn, eflux_errp = powerlaw.energy_flux_error(energy_min, energy_max)
    assert_allclose(eflux.value / 1e-12, 2.306978, rtol=7e-1)
    assert_allclose(eflux_errn.value / 1e-12, 0.819827, rtol=7e-1)
    assert_allclose(eflux_errp.value / 1e-12, 1.501977, rtol=7e-1)


def test_energy_flux_error_exp_cutoff_power_law():
    energy_min = 1 * u.TeV
    energy_max = 10 * u.TeV

    exppowerlaw = ExpCutoffPowerLawSpectralModel()
    exppowerlaw.parameters["index"].error = 0.4
    exppowerlaw.parameters["amplitude"].error = 1e-13
    exppowerlaw.parameters["lambda_"].error = 0.03

    eflux, eflux_errn, eflux_errp = exppowerlaw.energy_flux_error(
        energy_min, energy_max
    )

    assert_allclose(eflux.value / 1e-12, 2.751952, rtol=7e-1)
    assert_allclose(eflux_errn.value / 1e-12, 1.043413, rtol=7e-1)
    assert_allclose(eflux_errp.value / 1e-12, 2.052225, rtol=7e-1)


def test_integral_exp_cut_off_power_law_large_number_of_bins():
    energy = np.geomspace(1, 10, 100) * u.TeV
    energy_min = energy[:-1]
    energy_max = energy[1:]

    exppowerlaw = ExpCutoffPowerLawSpectralModel(
        amplitude="1e-11 TeV-1 cm-2 s-1", index=2
    )
    exppowerlaw.parameters["lambda_"].value = 1e-3
    powerlaw = PowerLawSpectralModel(amplitude="1e-11 TeV-1 cm-2 s-1", index=2)
    expected_flux = powerlaw.integral(energy_min, energy_max)

    flux = exppowerlaw.integral(energy_min, energy_max)

    assert_allclose(flux.value, expected_flux.value, rtol=0.01)


def test_template_ND(tmpdir, caplog):
    energy_axis = MapAxis.from_bounds(
        1.0, 100, 10, interp="log", name="energy_true", unit="GeV"
    )
    norm = MapAxis.from_bounds(0, 10, 10, interp="lin", name="norm", unit="")
    tilt = MapAxis.from_bounds(-1.0, 1, 5, interp="lin", name="tilt", unit="")
    region_map = RegionNDMap.create(
        region="icrs;point(83.63, 22.01)", axes=[energy_axis, norm, tilt]
    )
    region_map.data[:, :, :5, 0, 0] = 1
    region_map.data[:, :, 5:, 0, 0] = 2

    with caplog.at_level(logging.WARNING):
        template = TemplateNDSpectralModel(region_map)
        assert (
            'The filename is not defined. Therefore, the model will not be serialised correctly. To set the filename, the "template_model.filename" attribute can be used.'
            in [_.message for _ in caplog.records]
        )
    assert len(template.parameters) == 2
    assert template.parameters["norm"].value == 5
    assert template.parameters["tilt"].value == 0
    assert_allclose(template([1, 100, 1000] * u.GeV), [1.0, 2.0, 2.0])

    template.parameters["norm"].value = 1
    template.filename = str(tmpdir / "template_ND.fits")
    template.write()
    template.write(filename=str(tmpdir / "template_ND_2.fits"))

    dict_ = template.to_dict()
    template_new = TemplateNDSpectralModel.from_dict(dict_)
    assert_allclose(template_new.map.data, region_map.data)
    assert len(template_new.parameters) == 2
    assert template_new.parameters["norm"].value == 1
    assert template_new.parameters["tilt"].value == 0


def test_template_ND_no_energy(tmpdir):
    norm = MapAxis.from_bounds(0, 10, 10, interp="lin", name="norm", unit="")
    tilt = MapAxis.from_bounds(-1.0, 1, 5, interp="lin", name="tilt", unit="")
    region_map = RegionNDMap.create(
        region="icrs;point(83.63, 22.01)", axes=[norm, tilt]
    )
    region_map.data[:, :5, 0, 0] = 1
    region_map.data[:, 5:, 0, 0] = 2

    with pytest.raises(ValueError):
        TemplateNDSpectralModel(region_map)


@requires_data()
def test_template_ND_EBL(tmpdir):
    # TODO: add RegionNDMap.read(format="xspec")
    # Create EBL data array
    filename = "$GAMMAPY_DATA/ebl/ebl_franceschini.fits.gz"
    filename = make_path(filename)
    table_param = Table.read(filename, hdu="PARAMETERS")
    npar = len(table_param)
    par_axes = []
    idx_data = []
    for k in range(npar):
        name = table_param["NAME"][k].lower().strip()
        param, idx = np.unique(table_param[0]["VALUE"], return_index=True)
        par_axes.append(
            MapAxis(param, node_type="center", interp="lin", name=name, unit="")
        )
        idx_data.append(idx)
    idx_data.append(Ellipsis)
    idx_data = tuple(idx_data)

    # Get energy values
    table_energy = Table.read(filename, hdu="ENERGIES")
    energy_lo = u.Quantity(
        table_energy["ENERG_LO"], "keV", copy=COPY_IF_NEEDED
    )  # unit not stored in file
    energy_hi = u.Quantity(
        table_energy["ENERG_HI"], "keV", copy=COPY_IF_NEEDED
    )  # unit not stored in file
    energy = np.sqrt(energy_lo * energy_hi)

    # Get spectrum values
    table_spectra = Table.read(filename, hdu="SPECTRA")

    energy_axis = MapAxis(energy, node_type="center", interp="log", name="energy_true")
    region_map = RegionNDMap.create(
        region="galactic;point(0, 0)", axes=[energy_axis] + par_axes
    )
    # TODO: here we use a fake position, is it possible to allow region=None ?
    data = table_spectra["INTPSPEC"].data[idx_data]
    region_map.data[:, :, 0, 0] = data

    template = TemplateNDSpectralModel(region_map)
    assert len(template.parameters) == 1
    assert_allclose(template.parameters["redshift"].value, 1.001, rtol=1e-3)
    expected = [1.132092e00, 4.967878e-01, 1.596544e-06]
    assert_allclose(template([1, 100, 1000] * u.GeV), expected, rtol=1e-3)
    template.parameters["redshift"].value = 0.1
    template.filename = str(tmpdir / "template_ND_ebl_franceschini.fits")
    template.write()
    dict_ = template.to_dict()
    template_new = TemplateNDSpectralModel.from_dict(dict_)
    assert_allclose(template_new.map.data, region_map.data)
    assert len(template.parameters) == 1
    assert_allclose(template.parameters["redshift"].value, 0.1)


def test_incorrect_param_name():
    with pytest.raises(NameError):
        PowerLawSpectralModel(indxe=2)


def test_e_peak_super_4FGLDR3():
    model = SuperExpCutoffPowerLaw4FGLDR3SpectralModel()
    assert_quantity_allclose(model.e_peak, TEST_MODELS[9]["e_peak"], rtol=1e-2)

    model.index_1.value = 3
    model.index_2.value = 0.5
    model.expfactor.value = 0.5
    assert_quantity_allclose(model.e_peak, np.nan * u.TeV)

    model.index_2.value = 2
    model.expfactor.value = -1
    assert_quantity_allclose(model.e_peak, np.nan * u.TeV)

    model.index_2.value = -1
    assert_quantity_allclose(model.e_peak, np.nan * u.TeV)


def test_vectorized_integrate_spectrum():
    model = PowerLawSpectralModel()

    parameter_samples = [np.ones(10) * par.quantity for par in model.parameters]

    energy = [100, 1000, 10000] * u.GeV

    integral = integrate_spectrum(model, energy[:-1], energy[1:], ndecade=20)
    vector_integral = integrate_spectrum(
        model, energy[:-1], energy[1:], ndecade=20, parameter_samples=parameter_samples
    )

    assert integral.shape == (2,)
    assert_allclose(integral.to_value("cm-2s-1"), [9e-12, 9e-13])
    assert vector_integral.shape == (2, 10)
    assert_allclose(vector_integral[:, 0].to_value("cm-2s-1"), [9e-12, 9e-13])

    # check fail if model is not passed
    with pytest.raises(TypeError):
        integrate_spectrum(
            model.evaluate,
            energy[:-1],
            energy[1:],
            ndecade=20,
            parameter_samples=parameter_samples,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time


def _check_width(width):
    """Check and normalise width argument.

    Always returns tuple (lon, lat) as float in degrees.
    """
    if isinstance(width, tuple):
        lon = Angle(width[0], "deg").deg
        lat = Angle(width[1], "deg").deg
        return lon, lat
    else:
        angle = Angle(width, "deg").deg
        if np.isscalar(angle):
            return angle, angle
        else:
            return tuple(angle)


def _check_binsz(binsz):
    """Check and normalise bin size argument.

    Always returns an object with the same shape
    as the input where the spatial coordinates
    are a float in degrees.
    """
    if isinstance(binsz, tuple):
        lon_sz = Angle(binsz[0], "deg").deg
        lat_sz = Angle(binsz[1], "deg").deg
        return lon_sz, lat_sz
    elif isinstance(binsz, list):
        binsz[:2] = Angle(binsz[:2], unit="deg").deg
        return binsz
    return Angle(binsz, unit="deg").deg


def coordsys_to_frame(coordsys):
    if coordsys in ["CEL", "C"]:
        return "icrs"
    elif coordsys in ["GAL", "G"]:
        return "galactic"
    else:
        raise ValueError(f"Unknown coordinate system: '{coordsys}'")


def frame_to_coordsys(frame):
    if frame in ["icrs", "fk5", "fk4"]:
        return "CEL"
    elif frame in ["galactic"]:
        return "GAL"
    else:
        raise ValueError(f"Unknown coordinate frame '{frame}'")


class InvalidValue:
    """Class to define placeholder for invalid array values."""

    float = np.nan
    int = np.nan
    bool = np.nan
    time = Time(0.0, format="mjd", scale="tt")

    def __getitem__(self, dtype):
        if np.issubdtype(dtype, np.integer):
            return self.int
        elif np.issubdtype(dtype, np.floating):
            return self.float
        elif np.issubdtype(dtype, np.dtype(bool).type):
            return self.bool
        elif np.issubdtype(dtype, Time):
            return self.time
        else:
            raise ValueError(f"No invalid value placeholder defined for {dtype}")


class InvalidIndex:
    """Class to define placeholder for invalid array indices."""

    float = np.nan
    int = -1
    bool = False


INVALID_VALUE = InvalidValue()
INVALID_INDEX = InvalidIndex()


def edges_from_lo_hi(edges_lo, edges_hi):
    if np.isscalar(edges_lo.value) and np.isscalar(edges_hi.value):
        return u.Quantity([edges_lo, edges_hi])

    edges = edges_lo.copy()
    try:
        edges = edges.insert(len(edges), edges_hi[-1])
    except AttributeError:
        edges = np.insert(edges, len(edges), edges_hi[-1])
    return edges


def broadcast_axis_values_to_geom(geom, axis_name, use_center=True):
    """Reshape axis center array to the expected shape for broadcasting.

    Parameters
    ----------
    geom : `~gammapy.maps.Geom`
        the input Geom.
    axis_name : str
        input axis name. Must be part of the Geom non-spatial axes.
    use_centers : bool
        reshape center array if True, else use edges.

    Returns
    -------
    array : `~numpy.array` or `~astropy.Quantity`
        reshaped array.
    """
    if axis_name not in geom.axes.names:
        raise ValueError(f"Can't find axis {axis_name} in input Geom.")

    broadcast_shape = len(geom.data_shape) * [
        1,
    ]
    broadcast_shape[geom.axes.index_data(axis_name)] = -1

    if use_center:
        return geom.axes[axis_name].center.reshape(broadcast_shape)
    else:
        return geom.axes[axis_name].edges.reshape(broadcast_shape)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from regions import CircleSkyRegion
from gammapy.datasets import (
    Datasets,
    SpectrumDataset,
    SpectrumDatasetOnOff,
    FermipyDatasetsReader,
)
from gammapy.modeling.models import DatasetModels
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@requires_data()
def test_datasets_to_io(tmp_path):
    filedata = "$GAMMAPY_DATA/tests/models/gc_example_datasets.yaml"
    filemodel = "$GAMMAPY_DATA/tests/models/gc_example_models.yaml"

    datasets = Datasets.read(
        filename=filedata,
        filename_models=filemodel,
    )

    assert len(datasets) == 2
    assert len(datasets.models) == 5
    dataset0 = datasets[0]
    assert dataset0.name == "gc"
    assert dataset0.counts.data.sum() == 22258
    assert_allclose(dataset0.exposure.data.sum(), 8.057342e12, atol=0.1)
    assert dataset0.psf is not None
    assert dataset0.edisp is not None

    assert_allclose(dataset0.npred_background().data.sum(), 15726.8, atol=0.1)

    assert dataset0.background_model.name == "gc-bkg"

    dataset1 = datasets[1]
    assert dataset1.name == "g09"
    assert dataset1.background_model.name == "g09-bkg"

    assert (
        dataset0.models["gll_iem_v06_cutout"] == dataset1.models["gll_iem_v06_cutout"]
    )

    assert isinstance(dataset0.models, DatasetModels)
    assert len(dataset0.models) == 4
    assert dataset0.models[0].name == "gc"
    assert dataset0.models[1].name == "gll_iem_v06_cutout"
    assert dataset0.models[2].name == "gc-bkg"

    assert (
        dataset0.models["gc"].parameters["reference"]
        is dataset1.models["g09"].parameters["reference"]
    )
    assert_allclose(dataset1.models["g09"].parameters["lon_0"].value, 0.9, atol=0.1)

    datasets.write(
        filename=tmp_path / "written_datasets.yaml",
        filename_models=tmp_path / "written_models.yaml",
    )

    datasets.write(
        filename=tmp_path / "written_datasets.yaml",
        filename_models=tmp_path / "written_models.yaml",
        overwrite=True,
    )

    datasets_read = Datasets.read(
        filename=tmp_path / "written_datasets.yaml",
        filename_models=tmp_path / "written_models.yaml",
    )

    assert len(datasets.parameters) == 24

    assert len(datasets_read) == 2
    dataset0 = datasets_read[0]
    assert dataset0.name == "gc"
    assert dataset0.counts.data.sum() == 22258
    assert_allclose(dataset0.exposure.data.sum(), 8.057342e12, atol=0.1)
    assert dataset0.psf is not None
    assert dataset0.edisp is not None
    assert_allclose(dataset0.npred_background().data.sum(), 15726.8, atol=0.1)
    assert datasets[1].name == "g09"

    dataset_copy = dataset0.copy(name="dataset0-copy")
    assert dataset_copy.models is None


@requires_data()
def test_spectrum_datasets_to_io(tmp_path):
    filedata = "$GAMMAPY_DATA/tests/models/gc_example_datasets.yaml"
    filemodel = "$GAMMAPY_DATA/tests/models/gc_example_models.yaml"

    datasets = Datasets.read(
        filename=filedata,
        filename_models=filemodel,
    )
    reg = CircleSkyRegion(center=datasets[0]._geom.center_skydir, radius=1.0 * u.deg)
    dataset0 = datasets[0].to_spectrum_dataset(reg)
    datasets1 = Datasets([dataset0, datasets[1]])
    datasets1.write(
        filename=tmp_path / "written_datasets.yaml",
        filename_models=tmp_path / "written_models.yaml",
    )

    datasets_read = Datasets.read(
        filename=tmp_path / "written_datasets.yaml",
        filename_models=tmp_path / "written_models.yaml",
    )

    assert len(datasets_read.parameters) == 21

    assert len(datasets_read) == 2

    assert datasets_read[0].counts.data.sum() == 18429
    assert_allclose(datasets_read[0].exposure.data.sum(), 2.034089e10, atol=0.1)
    assert isinstance(datasets_read[0], SpectrumDataset)


@requires_data()
def test_ogip_writer(tmp_path):
    dataset = SpectrumDatasetOnOff.read(
        "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    )
    dataset.counts_off = None
    datasets = Datasets(dataset)

    datasets.write(tmp_path / "written_datasets.yaml")

    new_datasets = datasets.read(tmp_path / "written_datasets.yaml")

    assert new_datasets[0].counts_off is None


@requires_data()
def test_datasets_write_checksum(tmp_path):
    filedata = "$GAMMAPY_DATA/tests/models/gc_example_datasets.yaml"
    filemodel = "$GAMMAPY_DATA/tests/models/gc_example_models.yaml"

    datasets = Datasets.read(
        filename=filedata,
        filename_models=filemodel,
    )

    filename = tmp_path / "written_datasets.yaml"
    filename_models = tmp_path / "written_models.yaml"
    datasets.write(
        filename=filename,
        filename_models=filename_models,
        checksum=True,
    )

    assert "checksum" in filename.read_text()
    assert "checksum" in filename_models.read_text()

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        Datasets.read(filename=filename, filename_models=filename_models, checksum=True)

    # Remove checksum from datasets.yaml file
    yaml_content = filename.read_text()
    index = yaml_content.find("checksum")
    bad = make_path(tmp_path) / "bad_checksum.yaml"
    bad.write_text(yaml_content[:index])

    with pytest.warns(UserWarning):
        Datasets.read(filename=bad, filename_models=filename_models, checksum=True)

    # Modify models yaml file
    yaml_content = filename_models.read_text()
    yaml_content = yaml_content.replace("name: gc", "name: bad")
    bad = make_path(tmp_path) / "bad_models.yaml"
    bad.write_text(yaml_content)

    with pytest.warns(UserWarning):
        Datasets.read(filename=filename, filename_models=bad, checksum=True)


@requires_data()
def test_fermipy_datasets_reader():
    reader = FermipyDatasetsReader(
        "$GAMMAPY_DATA/tests/fermi/config_fermipy_std_5deg_qr.yaml", edisp_bins=1
    )
    datasets = reader.read()

    assert len(datasets) == 2
    assert datasets[0].counts.geom.axes[0].unit == "MeV"
    assert_allclose(datasets[0].background, 0)
    assert datasets[0].counts.geom.to_image() == datasets[0].exposure.geom.to_image()
    assert_allclose(datasets[0].exposure.data[0, 0, 0], 1.54938e11)
    assert_allclose(datasets[0].edisp.edisp_map.data[0, 0, 0, 0], 0.020409, rtol=1e-4)
    assert_allclose(
        datasets[0]._psf_kernel.psf_kernel_map.data.sum(axis=(1, 2)), 1, rtol=1e-5
    )
    assert_allclose(
        datasets[0].edisp.exposure_map.data, datasets[0].psf.exposure_map.data
    )
    assert datasets[0].name == "P8R3_SOURCEVETO_V3_PSF0_v1"
    assert datasets[1].name == "P8R3_SOURCEVETO_V3_PSF1_v1"
    assert datasets.models.names[0] == "isotropic_P8R3_SOURCEVETO_V3_PSF0_v1"
    assert datasets.models.names[1] == "isotropic_P8R3_SOURCEVETO_V3_PSF1_v1"

    path = make_path("$GAMMAPY_DATA/tests/fermi")
    dataset = reader.create_dataset(
        path / "ccube_00.fits",
        path / "bexpmap_00.fits",
        path / "psf_00.fits",
        path / "drm_00.fits",
    )
    assert not dataset.models

    reader = FermipyDatasetsReader(
        "$GAMMAPY_DATA/tests/fermi/config_fermipy_minimal.yaml", edisp_bins=1
    )

    datasets = reader.read()

    assert len(datasets) == 1
    assert datasets[0].counts.geom.axes[0].unit == "MeV"
    assert_allclose(datasets[0].background, 0)
    assert datasets[0].counts.geom.to_image() == datasets[0].exposure.geom.to_image()
    assert_allclose(datasets[0].exposure.data[0, 0, 0], 1.54938e11)
    assert_allclose(datasets[0].edisp.edisp_map.data[0, 0, 0, 0], 0.020409, rtol=1e-4)
    assert_allclose(
        datasets[0]._psf_kernel.psf_kernel_map.data.sum(axis=(1, 2)), 1, rtol=1e-5
    )
    assert_allclose(
        datasets[0].edisp.exposure_map.quantity, datasets[0].psf.exposure_map.quantity
    )
    assert datasets[0].edisp.exposure_map.unit == datasets[0].psf.exposure_map.unit
    assert datasets[0].edisp.exposure_map.unit == datasets[0].exposure.unit

    assert datasets[0].name == "P8R3_SOURCEVETO_V3_PSF1_v1"
    assert datasets.models.names[0] == "isotropic_P8R3_SOURCEVETO_V3_PSF1_v1"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import numpy as np
from .utils import get_random_state

__all__ = ["InverseCDFSampler"]


class InverseCDFSampler:
    """Inverse CDF sampler.

    It determines a set of random numbers and calculate the cumulative
    distribution function.

    Parameters
    ----------
    pdf : `~gammapy.maps.Map`
        Map of the predicted source counts.
    axis : int
        Axis along which sampling the indexes.
    random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}
        Defines random number generator initialisation.
        Passed to `~gammapy.utils.random.get_random_state`.
    """

    def __init__(self, pdf, axis=None, random_state=0):
        self.random_state = get_random_state(random_state)
        self.axis = axis

        if axis is not None:
            self.cdf = np.cumsum(pdf, axis=self.axis)
            self.cdf /= self.cdf[:, [-1]]
        else:
            self.pdf_shape = pdf.shape

            pdf = pdf.ravel() / pdf.sum()
            self.sortindex = np.argsort(pdf, axis=None)

            self.pdf = pdf[self.sortindex]
            self.cdf = np.cumsum(self.pdf)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def sample_axis(self):
        """Sample along a given axis.

        Returns
        -------
        index : tuple of `~numpy.ndarray`
            Coordinates of the drawn sample.
        """
        choices = self.random_state.uniform(high=1, size=len(self.cdf))
        shape_cdf = self.cdf.shape

        cdf_all = np.insert(self.cdf, 0, 0, axis=1)
        edges = np.arange(shape_cdf[1] + 1) - 0.5

        pix_coords = []

        for cdf, choice in zip(cdf_all, choices):
            pix = np.interp(choice, cdf, edges)
            pix_coords.append(pix)

        return np.array(pix_coords)

    def sample(self, size):
        """Draw sample from the given PDF.

        Parameters
        ----------
        size : int
            Number of samples to draw.

        Returns
        -------
        index : tuple of `~numpy.ndarray`
            Coordinates of the drawn sample.
        """
        # pick numbers which are uniformly random over the cumulative distribution function
        choice = self.random_state.uniform(high=1, size=size)

        # find the indices corresponding to this point on the CDF
        index = np.searchsorted(self.cdf, choice)
        index = self.sortindex[index]

        # map back to multi-dimensional indexing
        index = np.unravel_index(index, self.pdf_shape)
        index = np.vstack(index)

        index = index + self.random_state.uniform(low=-0.5, high=0.5, size=index.shape)
        return index



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy.io import fits
from astropy.table import Column, Table
from gammapy.utils.fits import earth_location_from_dict, earth_location_to_dict
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@pytest.fixture()
def header():
    filename = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )
    hdulist = fits.open(filename)
    return hdulist["EVENTS"].header


@pytest.fixture()
def table():
    t = Table(meta={"version": 42})
    t["a"] = np.array([1, 2], dtype=np.int32)
    t["b"] = Column(np.array([1, 2], dtype=np.int64), unit="m", description="Velocity")
    t["b"].meta["ucd"] = "spam"
    t["c"] = Column(["x", "yy"], "c")
    return t


def test_table_fits_io_astropy(table):
    """Test `astropy.table.Table` FITS I/O in Astropy.

    Having these tests in Gammapy is to check / ensure that the features
    we rely on work properly for all Astropy versions we support in CI
    (currently Astropy 1.3 and up)

    This is useful, because Table FITS I/O was pretty shaky for a while
    and incrementally improved over time.

    These are the same examples that we have in the docstring
    at the top of `gammapy/utils/fits.py`.
    """
    # Check Table -> BinTableHDU
    hdu = fits.BinTableHDU(table)
    assert hdu.header["TTYPE2"] == "b"
    assert hdu.header["TFORM2"] == "K"
    assert hdu.header["TUNIT2"] == "m"

    # Check BinTableHDU -> Table
    table2 = Table.read(hdu)
    assert isinstance(table2.meta, dict)
    assert table2.meta == {"VERSION": 42}
    assert table2["b"].unit == "m"
    # Note: description doesn't come back in older versions of Astropy
    # that we still support, so we're not asserting on that here for now.


@requires_data()
def test_earth_location_from_dict(header):
    location = earth_location_from_dict(header)

    assert_allclose(location.lon.value, 16.50022, rtol=1e-4)
    assert_allclose(location.lat.value, -23.271777, rtol=1e-4)
    assert_allclose(location.height.value, 1834.999999, rtol=1e-4)


@requires_data()
def test_earth_location_to_dict(header):
    location = earth_location_from_dict(header)
    loc_dict = earth_location_to_dict(location)

    assert_allclose(loc_dict["GEOLON"], 16.50022, rtol=1e-4)
    assert_allclose(loc_dict["GEOLAT"], -23.271777, rtol=1e-4)
    assert_allclose(loc_dict["ALTITUDE"], 1834.999999, rtol=1e-4)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.irf import EffectiveAreaTable2D, RadMax2D
from gammapy.maps import MapAxis
from gammapy.utils.compat import COPY_IF_NEEDED
from gammapy.utils.testing import mpl_plot_check


@pytest.fixture()
def rad_max_2d():
    n_energy = 10
    energy_axis = MapAxis.from_energy_bounds(
        50 * u.GeV, 100 * u.TeV, n_energy, name="energy"
    )

    n_offset = 5
    offset_axis = MapAxis.from_bounds(0, 2, n_offset, unit=u.deg, name="offset")

    shape = (n_energy, n_offset)
    rad_max = np.linspace(0.1, 0.5, n_energy * n_offset).reshape(shape)

    return RadMax2D(
        axes=[
            energy_axis,
            offset_axis,
        ],
        data=rad_max,
        unit=u.deg,
    )


def test_rad_max_roundtrip(rad_max_2d, tmp_path):
    rad_max_2d.write(tmp_path / "rad_max.fits")
    rad_max_read = RadMax2D.read(tmp_path / "rad_max.fits")

    assert_allclose(rad_max_read.data, rad_max_2d.data)
    assert rad_max_2d.axes == rad_max_read.axes


def test_rad_max_plot(rad_max_2d):
    with mpl_plot_check():
        rad_max_2d.plot_rad_max_vs_energy()


def test_rad_max_from_irf():
    e_bins = 3
    o_bins = 2
    energy_axis = MapAxis.from_energy_bounds(
        1 * u.TeV, 10 * u.TeV, nbin=e_bins, name="energy_true"
    )
    offset_axis = MapAxis.from_bounds(0 * u.deg, 3 * u.deg, nbin=o_bins, name="offset")
    aeff = EffectiveAreaTable2D(
        data=u.Quantity(np.ones((e_bins, o_bins)), u.m**2, copy=COPY_IF_NEEDED),
        axes=[energy_axis, offset_axis],
        is_pointlike=True,
    )

    with pytest.raises(ValueError):
        # not a point-like IRF
        RadMax2D.from_irf(aeff)

    with pytest.raises(ValueError):
        # missing rad_max
        RadMax2D.from_irf(aeff)

    aeff.meta["RAD_MAX"] = "0.2 deg"
    with pytest.raises(ValueError):
        # invalid format
        RadMax2D.from_irf(aeff)

    aeff.meta["RAD_MAX"] = 0.2
    rad_max = RadMax2D.from_irf(aeff)

    assert rad_max.axes["energy"].nbin == 1
    assert rad_max.axes["offset"].nbin == 1
    assert rad_max.axes["energy"].edges[0] == aeff.axes["energy_true"].edges[0]
    assert rad_max.axes["energy"].edges[1] == aeff.axes["energy_true"].edges[-1]
    assert rad_max.axes["offset"].edges[0] == aeff.axes["offset"].edges[0]
    assert rad_max.axes["offset"].edges[1] == aeff.axes["offset"].edges[-1]
    assert rad_max.quantity.shape == (1, 1)
    assert rad_max.quantity[0, 0] == 0.2 * u.deg


def test_rad_max_single_bin():
    energy_axis = MapAxis.from_energy_bounds(0.01, 100, 1, unit="TeV")
    offset_axis = MapAxis.from_bounds(
        0.0,
        5,
        1,
        unit="deg",
        name="offset",
    )

    rad_max = RadMax2D(
        data=[[0.1]] * u.deg,
        axes=[energy_axis, offset_axis],
        interp_kwargs={"method": "nearest", "fill_value": None},
    )

    value = rad_max.evaluate(energy=1 * u.TeV, offset=1 * u.deg)
    assert_allclose(value, 0.1 * u.deg)

    value = rad_max.evaluate(energy=[1, 2, 3] * u.TeV, offset=[[1]] * u.deg)
    assert value.shape == (1, 3)
    assert_allclose(value, 0.1 * u.deg)

    assert rad_max.is_fixed_rad_max



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Model parameter classes."""

import collections.abc
import copy
import html
import itertools
import logging
import numpy as np
from astropy import units as u
from astropy.table import Table
from gammapy.utils.interpolation import interpolation_scale
from gammapy.utils.scripts import make_name


__all__ = ["Parameter", "Parameters", "PriorParameter", "PriorParameters"]

log = logging.getLogger(__name__)


def _get_parameters_str(parameters):
    str_ = ""

    for par in parameters:
        if par.name == "amplitude":
            value_format, error_format = "{:10.2e}", "{:7.1e}"
        else:
            value_format, error_format = "{:10.3f}", "{:7.2f}"

        line = "\t{:21} {:8}: " + value_format + "\t {} {:<12s}\n"

        if par._link_label_io is not None:
            name = par._link_label_io
        else:
            name = par.name

        if par.frozen:
            frozen, error = "(frozen)", "\t\t"
        else:
            frozen = ""
            try:
                error = "+/- " + error_format.format(par.error)
            except AttributeError:
                error = ""
        str_ += line.format(name, frozen, par.value, error, par.unit)
    return str_.expandtabs(tabsize=2)


class Parameter:
    """A model parameter.

    Note that the parameter value has been split into
    a factor and scale like this::

        value = factor x scale

    Users should interact with the ``value``, ``quantity``
    or ``min`` and ``max`` properties and consider the fact
    that there is a ``factor`` and ``scale`` an implementation detail.

    That was introduced for numerical stability in parameter and error
    estimation methods, only in the Gammapy optimiser interface do we
    interact with the ``factor``, ``factor_min`` and ``factor_max`` properties,
    i.e. the optimiser "sees" the well-scaled problem.

    Parameters
    ----------
    name : str
        Name.
    value : float or `~astropy.units.Quantity`
        Value.
    scale : float, optional
        Scale (sometimes used in fitting).
    unit : `~astropy.units.Unit` or str, optional
        Unit. Default is "".
    min : float, str or `~astropy.units.quantity`, optional
        Minimum (sometimes used in fitting). If `None`, set to `numpy.nan`. Default is None.
    max : float, str or `~astropy.units.quantity`, optional
        Maximum (sometimes used in fitting). Default is `numpy.nan`.
    frozen : bool, optional
        Frozen (used in fitting).  Default is False.
    error : float, optional
        Parameter error. Default is 0.
    scan_min : float, optional
        Minimum value for the parameter scan. Overwrites scan_n_sigma.
        Default is None.
    scan_max : float, optional
        Maximum value for the parameter scan. Overwrites scan_n_sigma.
        Default is None.
    scan_n_values: int, optional
        Number of values to be used for the parameter scan. Default is 11.
    scan_n_sigma : int, optional
        Number of sigmas to scan. Default is 2.
    scan_values: `numpy.array`, optional
        Scan values. Overwrites all the scan keywords before.
        Default is None.
    scale_method : {'scale10', 'factor1', None}, optional
        Method used to set ``factor`` and ``scale``. Default is "scale10".
    interp : {"lin", "sqrt", "log"}, optional
        Parameter scaling to use for the scan. Default is "lin".
    prior : `~gammapy.modeling.models.Prior`, optional
        Prior set on the parameter. Default is None.
    """

    def __init__(
        self,
        name,
        value,
        unit="",
        scale=1,
        min=np.nan,
        max=np.nan,
        frozen=False,
        error=0,
        scan_min=None,
        scan_max=None,
        scan_n_values=11,
        scan_n_sigma=2,
        scan_values=None,
        scale_method="scale10",
        interp="lin",
        prior=None,
        scale_transform="lin",
    ):
        if not isinstance(name, str):
            raise TypeError(f"Name must be string, got '{type(name)}' instead")

        self._name = name
        self._link_label_io = None
        self._scale_method = scale_method
        self._scale_transform = scale_transform
        self.interp = interp
        self._scale = float(scale)
        self.frozen = frozen
        self._error = error
        self._type = None

        # TODO: move this to a setter method that can be called from `__set__` also!
        # Having it here is bad: behaviour not clear if Quantity and `unit` is passed.
        if isinstance(value, u.Quantity) or isinstance(value, str):
            val = u.Quantity(value)
            self.value = val.value
            self.unit = val.unit
        else:
            self.value = float(value)
            self.unit = unit

        self.min = min
        self.max = max
        self.scan_min = scan_min
        self.scan_max = scan_max
        self.scan_values = scan_values
        self.scan_n_values = scan_n_values
        self.scan_n_sigma = scan_n_sigma
        self.prior = prior

    def __get__(self, instance, owner):
        if instance is None:
            return self

        par = instance.__dict__[self.name]
        par._type = getattr(instance, "type", None)
        return par

    def __set__(self, instance, value):
        if isinstance(value, Parameter):
            instance.__dict__[self.name] = value
        else:
            par = instance.__dict__[self.name]
            raise TypeError(f"Cannot assign {value!r} to parameter {par!r}")

    def __set_name__(self, owner, name):
        if not self._name == name:
            raise ValueError(f"Expected parameter name '{name}', got {self._name}")

    @property
    def prior(self):
        """Prior applied to the parameter  as a `~gammapy.modeling.models.Prior`."""
        return self._prior

    @prior.setter
    def prior(self, value):
        if value is not None:
            from .models import Prior

            if isinstance(value, dict):
                from .models import Model

                self._prior = Model.from_dict({"prior": value})
            elif isinstance(value, Prior):
                self._prior = value
            else:
                raise TypeError(f"Invalid type: {value!r}")
        else:
            self._prior = value

    def prior_stat_sum(self):
        if self.prior is not None:
            return self.prior(self)

    @property
    def type(self):
        return self._type

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, value):
        self._error = self._set_quantity_str_float(value)

    @property
    def name(self):
        """Name as a string."""
        return self._name

    @property
    def factor(self):
        """Factor as a float."""
        return self._factor

    @factor.setter
    def factor(self, val):
        self._factor = float(val)
        self._value = float(self.inverse_transform(self._factor))

    @property
    def scale(self):
        """Scale as a float."""
        return self._scale

    @property
    def unit(self):
        """Unit as a `~astropy.units.Unit` object."""
        return self._unit

    @unit.setter
    def unit(self, val):
        self._unit = u.Unit(val)

    @property
    def min(self):
        """Minimum as a float."""
        return self._min

    @min.setter
    def min(self, val):
        """`~astropy.table.Table` has masked values for NaN. Replacing with NaN."""
        if isinstance(val, np.ma.core.MaskedConstant) or (val is None):
            self._min = np.nan
        else:
            self._min = self._set_quantity_str_float(val)

    @property
    def factor_min(self):
        """Factor minimum as a float (used by the optimizer).

        By default, when no transform is applied, ``factor_min = min / scale``,
        otherwise ``factor_min = transform(min)``.
        """
        return self.transform(self.min)

    @property
    def max(self):
        """Maximum as a float."""
        return self._max

    @max.setter
    def max(self, val):
        """`~astropy.table.Table` has masked values for NaN. Replacing with NaN."""
        if isinstance(val, np.ma.core.MaskedConstant) or (val is None):
            self._max = np.nan
        else:
            self._max = self._set_quantity_str_float(val)

    @property
    def factor_max(self):
        """Factor maximum as a float (used by the optimizer).

        By default, when no transform is applied, ``factor_max = max / scale``,
        otherwise ``factor_max = transform(max)``.
        """
        return self.transform(self.max)

    def _set_quantity_str_float(self, value):
        """Logics for min and max setter."""
        if isinstance(value, u.Quantity) or isinstance(value, str):
            value = u.Quantity(value)
            return float(value.to(self.unit).value)
        else:
            return float(value)

    def set_lim(self, min=None, max=None):
        """
        Set the min and/or max value for the parameter.

        Parameters
        ----------
        min, max: float, `~astropy.units.Quantity` or str, optional
            Minimum and Maximum value to assign to the parameter `min` and `max`.
            Default is None, which set `min` and `max` to `~numpy.nan`.
        """
        if min is not None:
            self.min = min
        if max is not None:
            self.max = max

    @property
    def scale_method(self):
        """Method used to set ``factor`` and ``scale``."""
        return self._scale_method

    @scale_method.setter
    def scale_method(self, val):
        if val not in ["scale10", "factor1"] and val is not None:
            raise ValueError(f"Invalid method: {val}")
        self.reset_autoscale()
        self._scale_method = val

    @property
    def scale_transform(self):
        """Scale interp : {"lin", "sqrt", "log"}."""
        return self._scale_transform

    @scale_transform.setter
    def scale_transform(self, val):
        if val not in ["lin", "log", "sqrt"]:
            raise ValueError(f"Invalid transform: {val}")
        self.reset_autoscale()
        self._scale_transform = val

    @property
    def frozen(self):
        """Frozen (used in fitting) (bool)."""
        return self._frozen

    @frozen.setter
    def frozen(self, val):
        if val in ["True", "False"]:
            val = bool(val)
        if not isinstance(val, bool) and not isinstance(val, np.bool_):
            raise TypeError(f"Invalid type: {val}, {type(val)}")
        self._frozen = val

    @property
    def value(self):
        """Value = factor x scale (float)."""
        return self._value

    @value.setter
    def value(self, val):
        self._value = float(val)
        self._factor = self.transform(val)

    @property
    def quantity(self):
        """Value times unit as a `~astropy.units.Quantity`."""
        return self.value * self.unit

    @quantity.setter
    def quantity(self, val):
        val = u.Quantity(val)

        if not val.unit.is_equivalent(self.unit):
            raise u.UnitConversionError(
                f"Unit must be equivalent to {self.unit} for parameter {self.name}"
            )

        self.value = val.value
        self.unit = val.unit

    @property
    def _step(self):
        return self.error if self.error > 0.0 else np.abs(self.value)

    # TODO: possibly allow to set this independently
    @property
    def conf_min(self):
        """Confidence minimum value as a `float`.

        Return parameter minimum if defined, otherwise  a default is estimated from value and error.
        """
        if not np.isnan(self.min):
            return self.min
        else:
            min_ = self.value - self._step * self.scan_n_sigma
            large_step = np.maximum(self._step, np.abs(self.value))
            min_ = np.minimum(min_, -large_step * 1e5)
            return min_

    # TODO: possibly allow to set this independently
    @property
    def conf_max(self):
        """Confidence maximum value as a `float`.

        Return parameter maximum if defined, otherwise a default is estimated from value and error.
        """
        if not np.isnan(self.max):
            return self.max
        else:
            max_ = self.value + self._step * self.scan_n_sigma
            large_step = np.maximum(self._step, np.abs(self.value))
            max_ = np.maximum(max_, large_step * 1e5)
            return max_

    @property
    def scan_min(self):
        """Stat scan minimum."""
        if self._scan_min is None:
            return self.value - self._step * self.scan_n_sigma

        return self._scan_min

    @property
    def scan_max(self):
        """Stat scan maximum."""
        if self._scan_max is None:
            return self.value + self._step * self.scan_n_sigma

        return self._scan_max

    @scan_min.setter
    def scan_min(self, value):
        """Stat scan minimum setter."""
        self._scan_min = value

    @scan_max.setter
    def scan_max(self, value):
        """Stat scan maximum setter."""
        self._scan_max = value

    @property
    def scan_n_sigma(self):
        """Stat scan n sigma."""
        return self._scan_n_sigma

    @scan_n_sigma.setter
    def scan_n_sigma(self, n_sigma):
        """Stat scan n sigma."""
        self._scan_n_sigma = int(n_sigma)

    @property
    def scan_values(self):
        """Stat scan values as a `numpy.ndarray`."""
        if self._scan_values is None:
            scale = interpolation_scale(self.interp)
            parmin, parmax = scale([self.scan_min, self.scan_max])
            values = np.linspace(parmin, parmax, self.scan_n_values)
            return scale.inverse(values)

        return self._scan_values

    @scan_values.setter
    def scan_values(self, values):
        """Set scan values."""
        self._scan_values = values

    def check_limits(self):
        """Emit a warning or error if value is outside the minimum/maximum range."""
        if not self.frozen:
            if (~np.isnan(self.min) and (self.value < self.min)) or (
                ~np.isnan(self.max) and (self.value > self.max)
            ):
                log.warning(
                    f"Value {self.value} is outside bounds [{self.min}, {self.max}]"
                    f" for parameter '{self.name}'"
                )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(name={self.name!r}, value={self.value!r}, "
            f"factor={self.factor!r}, scale={self.scale!r}, unit={self.unit!r}, "
            f"min={self.min!r}, max={self.max!r}, frozen={self.frozen!r}, prior={self.prior!r}, id={hex(id(self))})"
        )

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def copy(self):
        """Deep copy."""
        return copy.deepcopy(self)

    def update_from_dict(self, data):
        """Update parameters from a dictionary."""
        keys = ["value", "unit", "min", "max", "frozen", "prior"]
        for k in keys:
            if k == "prior" and data[k] == "":
                data[k] = None
            setattr(self, k, data[k])

    def to_dict(self):
        """Convert to dictionary."""
        output = {
            "name": self.name,
            "value": self.value,
            "unit": self.unit.to_string("fits"),
            "error": self.error,
            "min": self.min,
            "max": self.max,
            "frozen": self.frozen,
            "interp": self.interp,
            "scale_method": self.scale_method,
            "scale_transform": self.scale_transform,
        }

        if self._link_label_io is not None:
            output["link"] = self._link_label_io
        if self.prior is not None:
            output["prior"] = self.prior.to_dict()["prior"]
        return output

    def update_scale(self, value):
        """Update the parameter scale.

        Set ``factor`` and ``scale`` according to ``scale_method`` attribute.

        Available ``scale_method``.

        * ``scale10`` sets ``scale`` to power of 10,
          so that abs(factor) is in the range 1 to 10
        * ``factor1`` sets ``factor, scale = 1, value``

        In both cases the sign of value is stored in ``factor``,
        i.e. the ``scale`` is always positive.
        If ``scale_method`` is None the scaling is ignored.

        """
        if self.scale_method == "scale10":
            if value != 0:
                exponent = np.floor(np.log10(np.abs(value)))
                self._scale = np.power(10.0, exponent)

        elif self.scale_method == "factor1":
            self._scale = value

    def transform(self, value, update_scale=False):
        """Transform from value to factor (used by the optimizer).

        Parameters
        ----------
        value : float
            Parameter value.
        update_scale : bool, optional
            Update the scaling (used by the autoscale). Default is False.
        """
        interp_scale = interpolation_scale(self.scale_transform)
        transformed_value = interp_scale(value)
        if update_scale:
            self.update_scale(transformed_value)
        return transformed_value / self.scale

    def inverse_transform(self, factor):
        """Inverse transform from factor (used by the optimizer) to value.

        Parameters
        ----------
        factor : float
            Parameter factor.
        """
        interp_scale = interpolation_scale(self.scale_transform)
        value = interp_scale.inverse(self.scale * factor)
        return value

    def _inverse_transform_derivative(self, factor):
        """Inverse transform from factor (used by the optimizer) to value.

        Parameters
        ----------
        factor : float
            Parameter factor.
        """
        interp_scale = interpolation_scale(self.scale_transform)
        return interp_scale._inverse_deriv(self.scale * factor) * self.scale

    def autoscale(self):
        """Apply `~gammapy.utils.interpolation.interpolation_scale` and `scale_method` to the parameter."""
        self.factor = self.transform(self.value, update_scale=True)

    def reset_autoscale(self):
        """Reset scaling such as factor=value, scale=1."""
        self._factor = self._value
        self._scale = 1.0


class Parameters(collections.abc.Sequence):
    """Parameters container.

    - List of `Parameter` objects.
    - Covariance matrix.

    Parameters
    ----------
    parameters : list of `Parameter`
        List of parameters.
    """

    def __init__(self, parameters=None):
        if parameters is None:
            parameters = []
        else:
            parameters = list(parameters)

        self._parameters = parameters

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def check_limits(self):
        """Check parameter limits and emit a warning."""
        for par in self:
            par.check_limits()

    @property
    def prior(self):
        return [par.prior for par in self]

    def prior_stat_sum(self):
        parameters_stat_sum = 0
        for par in self:
            if par.prior is not None:
                parameters_stat_sum += par.prior_stat_sum()
        return parameters_stat_sum

    @property
    def types(self):
        """Parameter types."""
        return [par.type for par in self]

    @property
    def min(self):
        """Parameter minima as a `numpy.ndarray`."""
        return np.array([_.min for _ in self._parameters], dtype=np.float64)

    @min.setter
    def min(self, min_array):
        """Parameter minima as a `numpy.ndarray`."""
        if not len(self) == len(min_array):
            raise ValueError("Minima must have same length as parameter list")

        for min_, par in zip(min_array, self):
            par.min = min_

    @property
    def max(self):
        """Parameter maxima as a `numpy.ndarray`."""
        return np.array([_.max for _ in self._parameters], dtype=np.float64)

    @max.setter
    def max(self, max_array):
        """Parameter maxima as a `numpy.ndarray`."""
        if not len(self) == len(max_array):
            raise ValueError("Maxima must have same length as parameter list")

        for max_, par in zip(max_array, self):
            par.max = max_

    @property
    def value(self):
        """Parameter values as a `numpy.ndarray`."""
        return np.array([_.value for _ in self._parameters], dtype=np.float64)

    @value.setter
    def value(self, values):
        """Parameter values as a `numpy.ndarray`."""
        if not len(self) == len(values):
            raise ValueError("Values must have same length as parameter list")

        for value, par in zip(values, self):
            par.value = value

    @classmethod
    def from_stack(cls, parameters_list):
        """Create `Parameters` by stacking a list of other `Parameters` objects.

        Parameters
        ----------
        parameters_list : list of `Parameters`
            List of `Parameters` objects.
        """
        pars = itertools.chain(*parameters_list)
        return cls(pars)

    def copy(self):
        """Deep copy."""
        return copy.deepcopy(self)

    @property
    def free_parameters(self):
        """List of free parameters."""
        return self.__class__([par for par in self._parameters if not par.frozen])

    @property
    def unique_parameters(self):
        """Unique parameters as a `Parameters` object."""
        return self.__class__(dict.fromkeys(self._parameters))

    @property
    def free_unique_parameters(self):
        """List of free and unique parameters."""
        return self.__class__([par for par in self.unique_parameters if not par.frozen])

    @property
    def names(self):
        """List of parameter names."""
        return [par.name for par in self._parameters]

    def index(self, val):
        """Get position index for a given parameter.

        The input can be a parameter object, parameter name (str)
        or if a parameter index (int) is passed in, it is simply returned.
        """
        if isinstance(val, int):
            return val
        elif isinstance(val, Parameter):
            return self._parameters.index(val)
        elif isinstance(val, str):
            for idx, par in enumerate(self._parameters):
                if val == par.name:
                    return idx
            raise IndexError(f"No parameter: {val!r}")
        else:
            raise TypeError(f"Invalid type: {type(val)!r}")

    def __getitem__(self, key):
        """Access parameter by name, index or boolean mask."""
        if isinstance(key, np.ndarray) and key.dtype == bool:
            return self.__class__(list(np.array(self._parameters)[key]))
        else:
            idx = self.index(key)
            return self._parameters[idx]

    def __len__(self):
        return len(self._parameters)

    def __add__(self, other):
        if isinstance(other, Parameters):
            return Parameters.from_stack([self, other])
        else:
            raise TypeError(f"Invalid type: {other!r}")

    def to_dict(self):
        data = []

        for par in self._parameters:
            data.append(par.to_dict())

        return data

    @staticmethod
    def _create_default_table():
        name_to_type = {
            "type": "str",
            "name": "str",
            "value": "float",
            "unit": "str",
            "error": "float",
            "min": "float",
            "max": "float",
            "frozen": "bool",
            "link": "str",
            "prior": "str",
        }
        return Table(names=name_to_type.keys(), dtype=name_to_type.values())

    def update_link_label(self):
        """Update linked parameters labels used for serialisation and print."""
        params_list = []
        params_shared = []
        for param in self:
            if param not in params_list:
                params_list.append(param)
                params_list.append(param)
            elif param not in params_shared:
                params_shared.append(param)
        for param in params_shared:
            param._link_label_io = param.name + "@" + make_name()

    def to_table(self):
        """Convert parameter attributes to `~astropy.table.Table`."""
        self.update_link_label()
        table = self._create_default_table()

        for p in self._parameters:
            d = {k: v for k, v in p.to_dict().items() if k in table.colnames}
            if "prior" in d:
                d["prior"] = d["prior"]["type"]
            table.add_row(d)

        table["value"].format = ".4e"
        for name in ["error", "min", "max"]:
            table[name].format = ".3e"

        return table

    def __eq__(self, other):
        all_equal = np.all([p is p_new for p, p_new in zip(self, other)])
        return all_equal and len(self) == len(other)

    @classmethod
    def from_dict(cls, data):
        parameters = []

        for par in data:
            link_label = par.pop("link", None)
            par.pop("is_norm", None)
            parameter = Parameter(**par)
            parameter._link_label_io = link_label
            parameters.append(parameter)

        return cls(parameters=parameters)

    def set_parameter_factors(self, factors):
        """Set factor of all parameters.

        Used in the optimizer interface.
        """
        idx = 0
        for parameter in self._parameters:
            if not parameter.frozen:
                parameter.factor = factors[idx]
                idx += 1

    def autoscale(self):
        """Autoscale all parameters.

        See :func:`~gammapy.modeling.Parameter.autoscale`.

        """
        for par in self._parameters:
            par.autoscale()

    def select(
        self,
        name=None,
        type=None,
        frozen=None,
    ):
        """Create a mask of models, true if all conditions are verified.

        Parameters
        ----------
        name : str or list, optional
            Name of the parameter. Default is None.
        type : {None, "spatial", "spectral", "temporal"}
            Type of models. Default is None.
        frozen : bool, optional
            Select frozen parameters if True, exclude them if False. Default is None.

        Returns
        -------
        parameters : `Parameters`
           Selected parameters.
        """
        selection = np.ones(len(self), dtype=bool)

        if name and not isinstance(name, list):
            name = [name]

        for idx, par in enumerate(self):
            if name:
                selection[idx] &= np.any([_ == par.name for _ in name])

            if type:
                selection[idx] &= type == par.type

            if frozen is not None:
                if frozen:
                    selection[idx] &= par.frozen
                else:
                    selection[idx] &= ~par.frozen

        return self[selection]

    def freeze_all(self):
        """Freeze all parameters."""
        for par in self._parameters:
            par.frozen = True

    def unfreeze_all(self):
        """Unfreeze all parameters (even those frozen by default)."""
        for par in self._parameters:
            par.frozen = False

    def restore_status(self, restore_values=True):
        """Context manager to restore status.

        A copy of the values is made on enter,
        and those values are restored on exit.

        Parameters
        ----------
        restore_values : bool, optional
            Restore values if True, otherwise restore only frozen status. Default is None.

        Examples
        --------
        >>> from gammapy.modeling.models import PowerLawSpectralModel
        >>> pwl = PowerLawSpectralModel(index=2)
        >>> with pwl.parameters.restore_status():
        ...     pwl.parameters["index"].value = 3
        >>> print(pwl.parameters["index"].value) # doctest: +SKIP
        """
        return restore_parameters_status(self, restore_values)


class restore_parameters_status:
    def __init__(self, parameters, restore_values=True):
        self.restore_values = restore_values
        self._parameters = parameters
        self.values = [_.value for _ in parameters]
        self.frozen = [_.frozen for _ in parameters]

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        for value, par, frozen in zip(self.values, self._parameters, self.frozen):
            if self.restore_values:
                par.value = value
            par.frozen = frozen


class PriorParameter(Parameter):
    """Parameter of a `~gammapy.modeling.models.Prior`.

    A prior is a probability density function of a model parameter and can take different forms, including but not limited to Gaussian
    distributions and uniform distributions. The prior includes information or knowledge about the dataset or the
    parameters of the fit.

    Parameters
    ----------
    name : str
        Name.
    value : float or `~astropy.units.Quantity`
        Value.
    unit : `~astropy.units.Unit` or str, optional
        Unit. Default is "".

    Examples
    --------
    For a usage example see :doc:`/tutorials/details/priors` tutorial.
    """

    def __init__(
        self,
        name,
        value,
        unit="",
        scale=1,
        min=np.nan,
        max=np.nan,
        error=0,
        scale_method="scale10",
        scale_transform="lin",
    ):
        if not isinstance(name, str):
            raise TypeError(f"Name must be string, got '{type(name)}' instead")

        self._name = name
        self._scale_method = scale_method
        self._scale_transform = scale_transform
        self._scale = float(scale)
        self.min = min
        self.max = max
        self._error = error
        if isinstance(value, u.Quantity) or isinstance(value, str):
            val = u.Quantity(value)
            self.value = val.value
            self.unit = val.unit
        else:
            self.factor = value
            self.unit = unit

        self._type = "prior"

    def to_dict(self):
        """Convert to dictionary."""
        output = {
            "name": self.name,
            "value": self.value,
            "unit": self.unit.to_string("fits"),
            "error": self.error,
            "min": self.min,
            "max": self.max,
        }
        return output

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(name={self.name!r}, value={self.value!r}, "
            f"factor={self.factor!r}, scale={self.scale!r}, unit={self.unit!r}, "
            f"min={self.min!r}, max={self.max!r})"
        )


class PriorParameters(Parameters):
    """Container of parameter priors.

    - List of `PriorParameter` objects.

    Parameters
    ----------
    parameters : list of `PriorParameter`
        List of parameters.
    """

    def __init__(self, parameters=None):
        if parameters is None:
            parameters = []
        else:
            parameters = list(parameters)

        self._parameters = parameters

    def to_table(self):
        """Convert parameter attributes to `~astropy.table.Table`."""
        rows = []
        for p in self._parameters:
            d = p.to_dict()
            rows.append({**dict(type=p.type), **d})
        table = Table(rows)

        table["value"].format = ".4e"
        for name in ["error", "min", "max"]:
            table[name].format = ".3e"

        return table

    @classmethod
    def from_dict(cls, data):
        parameters = []

        for par in data:
            parameter = PriorParameter(**par)
            parameters.append(parameter)

        return cls(parameters=parameters)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Multi-Gaussian distribution utilities (Gammapy internal)."""
import html
import numpy as np
from astropy import units as u
from gammapy.utils.roots import find_roots


class Gauss2DPDF:
    """2D symmetric Gaussian PDF.

    Reference: http://en.wikipedia.org/wiki/Multivariate_normal_distribution#Bivariate_case

    Parameters
    ----------
    sigma : float
        Gaussian width.
    """

    def __init__(self, sigma=1):
        self.sigma = sigma

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def _sigma2(self):
        """Sigma squared as a float."""
        return self.sigma * self.sigma

    @property
    def amplitude(self):
        """PDF amplitude at the center as a float."""
        return self.__call(0, 0)

    def __call__(self, x, y=0):
        """dp / (dx dy) at position (x, y).

        Parameters
        ----------
        x : `~numpy.ndarray`
            x coordinate.
        y : `~numpy.ndarray`, optional
            y coordinate.

        Returns
        -------
        dpdxdy : `~numpy.ndarray`
            dp / (dx dy).
        """
        theta2 = x * x + y * y
        amplitude = 1 / (2 * np.pi * self._sigma2)
        exponent = -0.5 * theta2 / self._sigma2
        return amplitude * np.exp(exponent)

    def dpdtheta2(self, theta2):
        """dp / dtheta2 at position theta2 = theta ^ 2.

        Parameters
        ----------
        theta2 : `~numpy.ndarray`
            Offset squared.

        Returns
        -------
        dpdtheta2 : `~numpy.ndarray`
            dp / dtheta2.
        """
        amplitude = 1 / (2 * self._sigma2)
        exponent = -0.5 * theta2 / self._sigma2
        return amplitude * np.exp(exponent)

    def containment_fraction(self, rad):
        """Containment fraction.

        Parameters
        ----------
        rad : `~numpy.ndarray`
            Offset.

        Returns
        -------
        containment_fraction : `~numpy.ndarray`
            Containment fraction.
        """
        return 1 - np.exp(-0.5 * rad**2 / self._sigma2)

    def containment_radius(self, containment_fraction):
        """Containment angle for a given containment fraction.

        Parameters
        ----------
        containment_fraction : `~numpy.ndarray`
            Containment fraction.

        Returns
        -------
        containment_radius : `~numpy.ndarray`
            Containment radius.
        """
        return self.sigma * np.sqrt(-2 * np.log(1 - containment_fraction))

    def gauss_convolve(self, sigma):
        """Convolve with another Gaussian 2D PDF.

        Parameters
        ----------
        sigma : `~numpy.ndarray` or float
            Gaussian width of the new Gaussian 2D PDF to convolve with.

        Returns
        -------
        gauss_convolve : `~gammapy.modeling.models.Gauss2DPDF`
            Convolution of both Gaussians.
        """
        new_sigma = np.sqrt(self._sigma2 + sigma**2)
        return Gauss2DPDF(new_sigma)


class MultiGauss2D:
    """Sum of multiple 2D Gaussians.

    Parameters
    ----------
    sigmas : `~numpy.ndarray`
            Widths of the Gaussians to add.
    norms : `~numpy.ndarray`, optional
            Normalizations of the Gaussians to add.

    Notes
    -----
    * This sum is no longer a PDF, it is not normalized to 1.
    * The "norm" of each component represents the 2D integral,
      not the amplitude at the origin.
    """

    def __init__(self, sigmas, norms=None):
        # If no norms are given, you have a PDF.
        self.components = [Gauss2DPDF(sigma) for sigma in sigmas]

        if norms is None:
            self.norms = np.ones(len(self.components))
        else:
            self.norms = norms

    def __call__(self, x, y=0):
        """dp / (dx dy) at position (x, y).

        Parameters
        ----------
        x : `~numpy.ndarray`
            x coordinate.
        y : `~numpy.ndarray`, optional
            y coordinate. Default is 0.

        Returns
        -------
        total : `~numpy.ndarray`
            dp / (dx dy).
        """
        values = []
        for norm, component in zip(self.norms, self.components):
            values.append(norm * component(x, y))

        return np.stack(values).sum(axis=0)

    @property
    def n_components(self):
        """Number of components as an integer."""
        return len(self.components)

    @property
    def sigmas(self):
        """Array of Gaussian widths as an `~numpy.ndarray`."""
        return u.Quantity([_.sigma for _ in self.components])

    @property
    def integral(self):
        """Integral as sum of norms as an `~numpy.ndarray`."""
        return np.nansum(self.norms, axis=0)

    @property
    def amplitude(self):
        """Amplitude at the center as a float."""
        return self.__call__(0, 0)

    @property
    def max_sigma(self):
        """Largest Gaussian width as a float."""
        return self.sigmas.max()

    @property
    def eff_sigma(self):
        r"""Effective Gaussian width for single-Gauss approximation as a float.

        Notes
        -----
        The effective Gaussian width is given by:

        .. math:: \sigma_\mathrm{eff} = \sqrt{\sum_i N_i \sigma_i^2}

        where ``N`` is normalization and ``sigma`` is width.

        """
        sigma2s = [component._sigma2 for component in self.components]
        return np.sqrt(np.sum(self.norms * sigma2s))

    def dpdtheta2(self, theta2):
        """dp / dtheta2 at position theta2 = theta ^ 2.

        Parameters
        ----------
        theta2 : `~numpy.ndarray`
            Offset squared.

        Returns
        -------
        dpdtheta2 : `~numpy.ndarray`
            dp / dtheta2.
        """
        values = []
        # Actually this is only a PDF if sum(norms) == 1
        for norm, component in zip(self.norms, self.components):
            values.append(norm * component.dpdtheta2(theta2))
        return np.sum(values, axis=0)

    def normalize(self):
        """Normalize function.

        Returns
        -------
        norm_multigauss : `~gammapy.modeling.models.MultiGauss2D`
           Normalized function.
        """
        with np.errstate(divide="ignore", invalid="ignore"):
            self.norms = np.nan_to_num(self.norms / self.integral)

    def containment_fraction(self, rad):
        """Containment fraction.

        Parameters
        ----------
        rad : `~numpy.ndarray`
            Offset.

        Returns
        -------
        containment_fraction : `~numpy.ndarray`
            Containment fraction.
        """
        values = []

        for norm, component in zip(self.norms, self.components):
            values.append(norm * component.containment_fraction(rad))

        return np.sum(values, axis=0)

    def containment_radius(self, containment_fraction):
        """Containment angle for a given containment fraction.

        Parameters
        ----------
        containment_fraction : `~numpy.ndarray`
            Containment fraction.

        Returns
        -------
        containment_radius : `~numpy.ndarray`
            Containment radius.
        """
        rad_max = 1e3

        def f(rad):
            # positive if theta too large
            return self.containment_fraction(rad * u.deg) - containment_fraction

        roots, res = find_roots(f, lower_bound=0, upper_bound=rad_max, nbin=1)
        if np.isnan(roots[0]) or np.allclose(roots[0], rad_max):
            rad = np.inf
        else:
            rad = roots[0]
        return rad * u.deg

    def match_sigma(self, containment_fraction):
        """Compute equivalent Gauss width.

        Find the sigma of a single-Gaussian distribution that
        approximates this one, such that theta matches for a given
        containment.

        Parameters
        ----------
        containment_fraction : `~numpy.ndarray`
            Containment fraction.

        Returns
        -------
        sigma : `~numpy.ndarray`
            Equivalent containment radius.
        """
        theta1 = self.containment_radius(containment_fraction)
        theta2 = Gauss2DPDF(sigma=1).containment_radius(containment_fraction)
        return theta1 / theta2

    def gauss_convolve(self, sigma, norm=1):
        """Convolve with another Gaussian.

        Compute new norms and sigmas of all the components such that
        the new distribution represents the convolved old distribution
        by a Gaussian of width sigma and then multiplied by norm.

        This MultiGauss2D is unchanged, a new one is created and returned.
        This is useful if you need to e.g. compute theta for one PSF
        and many sigmas.

        Parameters
        ----------
        sigma : `~numpy.ndarray` or float
            Gaussian width of the new Gaussian 2D PDF to convolve with.
        norm : `~numpy.ndarray` or float
            Normalization of the new Gaussian 2D PDF to convolve with.

        Returns
        -------
        new_multi_gauss_2d : `~gammapy.modeling.models.MultiGauss2D`
            Convolution as new MultiGauss2D.
        """
        sigmas, norms = [], []
        for ii in range(self.n_components):
            sigmas.append(self.components[ii].gauss_convolve(sigma).sigma)
            norms.append(self.norms[ii] * norm)

        return MultiGauss2D(sigmas, norms)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from pydantic import ValidationError
from gammapy.utils.metadata import CreatorMetaData, TargetMetaData
from gammapy.utils.testing import requires_data
from gammapy.version import version
from ..metadata import FluxMetaData


@pytest.fixture()
def default():
    default = FluxMetaData(
        sed_type="likelihood",
        creation=CreatorMetaData(
            date="2022-01-01", creator="gammapy test", origin="CTA"
        ),
        target=TargetMetaData(
            name="PKS2155-304",
            position=SkyCoord(
                "21 58 52.06 -30 13 32.11", frame="icrs", unit=(u.hourangle, u.deg)
            ),
        ),
        n_sigma=2.0,
        n_sigma_ul=None,
    )
    return default


@requires_data()
def test_creator(default):

    assert default.creation.creator == "gammapy test"
    assert default.sed_type == "likelihood"
    assert default.sed_type_init is None
    assert default.target.name == "PKS2155-304"
    assert default.target.position is not None

    with pytest.raises(ValidationError):
        default.target.position = 2.0

    with pytest.raises(ValidationError):
        default.sed_type = "test"

    default = FluxMetaData()
    assert default.creation.creator == f"Gammapy {version}"


def test_from_header():
    tdict = {
        "SED_TYPE": "dnde",
        "N_SIGMA": "4.3",
        "OBJECT": "RXJ",
        "RA_OBJ": "123.5",
        "DEC_OBJ": "-4.8",
        "OBS_IDS": ["1", "2", "3", "6"],
        "DATASETS": "myanalysis",
        "INSTRU": "HAWC",
        "CREATED": "2023-10-27 16:05:09.795",
        "ORIGIN": "Gammapy v2.3",
        "CREATOR": "mynotebook",
        "EPHEM": "fermi23.1",
    }

    meta = FluxMetaData.from_header(tdict)
    assert meta.n_sigma == 4.3
    assert meta.creation.origin == "Gammapy v2.3"
    assert meta.target.position == SkyCoord(123.5 * u.deg, -4.8 * u.deg, frame="icrs")
    assert meta.optional is None
    # TODO: add for support: assert meta.optional["EPHEM"] == "fermi23.1"


@requires_data()
def test_to_header(default):
    hdr = default.to_header()
    assert hdr["OBJECT"] == "PKS2155-304"
    assert hdr["ORIGIN"] == "CTA"
    assert hdr["RA_OBJ"] == 329.7169166666666
    assert hdr["N_SIGMA"] == 2.0
    assert hdr["CREATOR"] == "gammapy test"
